"""Primary pipeline to process and evaluate data to be imported
into the MySQL database."""

import argparse
from datetime import datetime, date
import logging
import os
import pathlib
import shutil
import sys

from tabulate import tabulate

import pdm_utils     # to get version number.
from pdm_utils.classes.alchemyhandler import AlchemyHandler
from pdm_utils.classes import bundle
from pdm_utils.classes import genomepair
from pdm_utils.constants import constants, eval_descriptions
from pdm_utils.functions import basic
from pdm_utils.functions import configfile
from pdm_utils.functions import tickets
from pdm_utils.functions import fileio
from pdm_utils.functions import flat_files
from pdm_utils.functions import phagesdb
from pdm_utils.functions import mysqldb
from pdm_utils.functions import mysqldb_basic
from pdm_utils.functions import eval_modes

# Add a logger named after this module. Then add a null handler, which
# suppresses any output statements. This allows other modules that call this
# module to define the handler and output formats. If this module is the
# main module being called, the top level main function configures
# the root logger and primary file handle.
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


DEFAULT_OUTPUT_FOLDER = os.getcwd()
IMPORT_DATE = datetime.today().replace(hour=0, minute=0, second=0,
                                       microsecond=0)
CURRENT_DATE = date.today().strftime("%Y%m%d")
RESULTS_FOLDER = f"{CURRENT_DATE}_import"
SUCCESS_FOLDER = "success"
FAIL_FOLDER = "fail"
GENOMES_FOLDER = "genomes"
LOGS_FOLDER = "logs"
VERSION = pdm_utils.__version__
EDD = eval_descriptions.EVAL_DESCRIPTIONS
MAIN_LOG_FILE = "import.log"


def main(unparsed_args_list):
    """Runs the complete import pipeline.

    This is the only function of the pipeline that requires user input.
    All other functions can be implemented from other scripts.

    :param unparsed_args_list:
        List of strings representing command line arguments.
    :type unparsed_args_list: list
    """

    args = parse_args(unparsed_args_list)

    # Validate folders and files.
    args.input_folder = basic.set_path(args.input_folder, kind="dir",
                                       expect=True)
    args.output_folder = basic.set_path(args.output_folder, kind="dir",
                                        expect=True)
    args.import_table = basic.set_path(args.import_table, kind="file",
                                       expect=True)

    results_folder = pathlib.Path(RESULTS_FOLDER)
    results_path = basic.make_new_dir(args.output_folder,
                                      results_folder, attempt=50)

    if results_path is None:
        print("Unable to create results folder.")
        sys.exit(1)

    # Create config object with data obtained from file and/or defaults.
    config = configfile.build_complete_config(args.config_file)
    mysql_creds = config["mysql"]

    log_file = pathlib.Path(results_path, MAIN_LOG_FILE)

    # Set up root logger.
    logging.basicConfig(filename=log_file, filemode="w",
                        level=logging.DEBUG,
                        format="pdm_utils import: %(levelname)s: %(message)s")
                        # format="%(name)s - %(levelname)s - %(message)s")

    logger.info(f"pdm_utils version: {VERSION}")
    logger.info(f"Import run date: {CURRENT_DATE}")
    logger.info(f"Command line arguments: {' '.join(unparsed_args_list)}")
    logger.info(f"Results directory: {results_path}")
    logger.info("Command line arguments verified.")

    # Verify database connection and schema compatibility.
    alchemist = AlchemyHandler(database=args.database,
                               username=mysql_creds["user"],
                               password=mysql_creds["password"])
    alchemist.connect(login_attempts=5, pipeline=True)
    engine = alchemist.engine
    logger.info(f"Connected to database: {args.database}.")
    mysqldb.check_schema_compatibility(engine, "the import pipeline")
    logger.info(f"Schema version is compatible.")

    # If everything checks out, pass on args for data input/output.
    data_io(engine=engine,
            genome_folder=args.input_folder,
            import_table_file=args.import_table,
            genome_id_field=args.genome_id_field,
            host_genus_field=args.host_genus_field,
            prod_run=args.prod_run,
            description_field=args.description_field,
            eval_mode=args.eval_mode,
            output_folder=results_path,
            interactive=args.interactive, accept_warning=args.accept_warning)

    logger.info("Import complete.")


def parse_args(unparsed_args_list):
    """Verify the correct arguments are selected for import new genomes.

    :param unparsed_args_list:
        List of strings representing command line arguments.
    :type unparsed_args_list: list
    :returns: ArgumentParser Namespace object containing the parsed args.
    :rtype: Namespace
    """

    import_help = (
        "Pipeline to import new genome data into a MySQL database.")
    database_help = (
        "Name of the MySQL database to import the genomes.")
    input_folder_help = (
        "Path to the folder containing GenBank-formatted "
        "flat files to be processed.")
    output_folder_help = (
        "Path to the folder to store results. "
        f"Default is {DEFAULT_OUTPUT_FOLDER}")
    import_table_help = (
        "Path to the CSV-formatted table containing "
        "instructions to process each genome.")
    genome_id_field_help = (
        "Indicates the flat file field that should be used "
        "as the unique identifier for the genome during import.")
    host_genus_field_help = (
        "Indicates the flat file field that should be used "
        "to identify the host genus for the genome during import.")
    prod_run_help = (
        "Indicates whether the script should make any changes to the database. "
        "If True, the production run will implement all changes in the "
        "indicated database. If False, the test run will not "
        "implement any changes.")
    eval_mode_help = (
        "Indicates the evaluation configuration for importing genomes.")
    description_field_help = (
        "Indicates the field in CDS features that is expected "
        "to store the gene description.")
    interactive_help = (
        "Indicates whether interactive evaluation of data is permitted.")
    accept_warning_help = (
        "Indicates whether warnings are accepted without interactivity.")
    config_file_help = "Path to the file containing user-specific login data."

    parser = argparse.ArgumentParser(description=import_help)
    parser.add_argument("database", type=str, help=database_help)
    parser.add_argument("input_folder", type=pathlib.Path,
        help=input_folder_help)
    parser.add_argument("import_table", type=pathlib.Path,
        help=import_table_help)
    parser.add_argument("-g", "--genome_id_field", type=str.lower,
        default="_organism_name",
        choices=["_organism_name", "_description_name",
                 "_source_name", "filename"],
        help=genome_id_field_help)
    parser.add_argument("-hg", "--host_genus_field", type=str.lower,
        default="_organism_host_genus",
        choices=["_organism_host_genus", "_description_host_genus",
                 "_source_host_genus"],
        help=host_genus_field_help)
    parser.add_argument("-p", "--prod_run", action="store_true",
        default=False, help=prod_run_help)
    parser.add_argument("-e", "--eval_mode", type=str.lower,
        choices=list(eval_modes.EVAL_MODES.keys()), default="final",
        help=eval_mode_help)
    parser.add_argument("-d", "--description_field", type=str.lower,
        default="product", choices=list(constants.DESCRIPTION_FIELD_SET),
        help=description_field_help)
    parser.add_argument("-o", "--output_folder", type=pathlib.Path,
        default=pathlib.Path(DEFAULT_OUTPUT_FOLDER), help=output_folder_help)
    parser.add_argument("-i", "--interactive", action="store_true",
        default=False, help=interactive_help)
    parser.add_argument("-Y", "--accept_warning", action="store_true",
                        default=False, help=accept_warning_help)
    parser.add_argument("-c", "--config_file", type=pathlib.Path,
                        help=config_file_help, default=None)


    # Assumed command line arg structure:
    # python3 -m pdm_utils.run <pipeline> <additional args...>
    # sys.argv:      [0]            [1]         [2...]
    args = parser.parse_args(unparsed_args_list[2:])

    return args


def data_io(engine=None, genome_folder=pathlib.Path(),
            import_table_file=pathlib.Path(), genome_id_field="",
            host_genus_field="", prod_run=False, description_field="",
            eval_mode="", output_folder=pathlib.Path(), interactive=False,
            accept_warning=False):
    """Set up output directories, log files, etc. for import.

    :param engine:
        SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param genome_folder: Path to the folder of flat files.
    :type genome_folder: Path
    :param import_table_file: Path to the import table file.
    :type import_table_file: Path
    :param genome_id_field:
        The SeqRecord attribute that stores the genome identifier/name.
    :type genome_id_field: str
    :param host_genus_field:
        The SeqRecord attribute that stores the host genus identifier/name.
    :type host_genus_field: str
    :param prod_run: Indicates whether MySQL statements will be executed.
    :type prod_run: bool
    :param description_field:
        The SeqFeature attribute that stores the feature's description.
    :type description_field: str
    :param eval_mode:
        Name of the evaluation mode to evaluation genomes.
    :type eval_mode: str
    :param output_folder: Path to the folder to store results.
    :type output_folder: Path
    :param interactive:
        Indicates whether user is able to interact with genome evaluations
        at run time
    :type interactive: bool
    :param accept_warning:
        Toggles whether the import pipeline will accept warnings without
        interactivity.
    :type accept_warning: bool
    """

    logger.info("Setting up environment.")
    # Get the files to process.
    files_to_process = basic.identify_contents(genome_folder, kind="file",
                                               ignore_set=set([".DS_Store"]))
    file_msg = f"There are {len(files_to_process)} flat files to evaluate."
    if len(files_to_process) == 0:
        logger.error(file_msg)
        print(file_msg)
        sys.exit(1)
    else:
        log_and_print(file_msg, True)

    # Get the tickets.
    eval_flags = eval_modes.get_eval_flag_dict(eval_mode.lower())
    eval_data_dict = {"eval_mode": eval_mode, "eval_flag_dict": eval_flags}
    ticket_dict = prepare_tickets(import_table_file,
                                  eval_data_dict,
                                  description_field,
                                  constants.IMPORT_TABLE_STRUCTURE)
    if ticket_dict is None:
        tkt_msg = "Invalid import table. Unable to evaluate flat files."
        logger.error(tkt_msg)
        print(f"\n\n\nError: {tkt_msg}")
        sys.exit(1)
    else:
        progress = f"There are {len(ticket_dict.keys())} import ticket(s)."
        log_and_print(progress, True)

    # Set up output folders. Creating all folders in advance, since
    # logging the flatfile log files need a valid path.
    success_path = pathlib.Path(output_folder, SUCCESS_FOLDER)
    success_genomes_path = pathlib.Path(success_path, GENOMES_FOLDER)
    success_logs_path = pathlib.Path(success_path, LOGS_FOLDER)

    failed_path = pathlib.Path(output_folder, FAIL_FOLDER)
    failed_genomes_path = pathlib.Path(failed_path, GENOMES_FOLDER)
    failed_logs_path = pathlib.Path(failed_path, LOGS_FOLDER)

    output_folders = [success_path, success_genomes_path, success_logs_path,
                      failed_path, failed_genomes_path, failed_logs_path]
    for folder in output_folders:
        folder.mkdir()
    log_folder_paths_dict = {SUCCESS_FOLDER: success_logs_path,
                             FAIL_FOLDER: failed_logs_path}
    start_count = mysqldb_basic.get_table_count(engine, "phage")

    # Evaluate files and tickets.
    results_tuple = process_files_and_tickets(
                        ticket_dict, files_to_process,
                        engine=engine,
                        prod_run=prod_run,
                        genome_id_field=genome_id_field,
                        host_genus_field=host_genus_field,
                        interactive=interactive, accept_warning=accept_warning,
                        log_folder_paths_dict=log_folder_paths_dict)
    success_ticket_list = results_tuple[0]
    failed_ticket_list = results_tuple[1]
    success_filepath_list = results_tuple[2]
    failed_filepath_list = results_tuple[3]
    evaluation_dict = results_tuple[4]

    final_count = mysqldb_basic.get_table_count(engine, "phage")

    # Output data.
    logger.info("Logging successful tickets and files.")
    summary = [
        "Summary of import: ",
        f"{start_count} genome(s) in the database before import.",
        f"{final_count} genome(s) in the database after import.",
        f"{len(success_ticket_list)} ticket(s) successfully processed.",
        f"{len(success_filepath_list)} genome(s) successfully processed."
        ]

    headers = constants.IMPORT_TABLE_STRUCTURE["order"]
    if (len(success_ticket_list) > 0 or len(success_filepath_list) > 0):
        if len(success_ticket_list) > 0:
            success_tkt_file = pathlib.Path(success_path, "import_table.csv")
            fileio.export_data_dict(success_ticket_list, success_tkt_file,
                                       headers, include_headers=True)
        if len(success_filepath_list) > 0:
            for file in success_filepath_list:
                new_file = pathlib.Path(success_genomes_path, file.name)
                shutil.copy(str(file), str(new_file))

    logger.info("Logging failed tickets and files.")
    if (len(failed_ticket_list) > 0 or len(failed_filepath_list) > 0):
        if len(failed_ticket_list) > 0:
            failed_tkt_file = pathlib.Path(failed_path, "import_table.csv")
            fileio.export_data_dict(failed_ticket_list, failed_tkt_file,
                                       headers, include_headers=True)
            summary.append(f"{len(failed_ticket_list)} ticket(s) NOT implemented:")
            for tkt in failed_ticket_list:
                tkt_summary = (f"Ticket Type: {tkt['type']}. "
                               f"PhageID: {tkt['phage_id']}.")
                summary.append(f"{tkt_summary}")

        if len(failed_filepath_list) > 0:
            for file in failed_filepath_list:
                new_file = pathlib.Path(failed_genomes_path, file.name)
                shutil.copy(str(file), str(new_file))
            summary.append(f"{len(failed_filepath_list)} genome(s) NOT imported:")
            for file in failed_filepath_list:
                summary.append(f"{file.name}")

    # Now check which folders are empty and remove them.
    # List needs to be reversed. For instance, the main 'success' folder
    # can't be removed until the 'genomes' and 'logs' folders are removed.
    for folder in reversed(output_folders):
        if len(basic.identify_contents(folder, kind=None)) == 0:
            folder.rmdir()

    print("\n\n\n")
    for line in summary:
        print(line)
        logger.info(line)

    # Close all connections in the connection pool.
    engine.dispose()


def log_evaluations(dict_of_dict_of_lists, logfile_path=None):
    """Export evaluations to log.

    :param dict_of_dict_of_lists:
        Dictionary of evaluation dictionaries.
        Key1 = Bundle ID.
        Value1 = dictionary for each object in the Bundle.
        Key2 = object type ('bundle', 'ticket', etc.)
        Value2 = List of evaluation objects.
    :type dict_of_dict_of_lists: dict
    :param logfile_path: Path to the log file.
    :type logfile_path: Path
    """
    # Example structure:
    #
    #     {1: {"bundle": [eval_object1, ...],
    #          "ticket": [eval_object1, ...],
    #          "genome": [eval_object1, ...]},
    #      2: {...}}

    # Create a new logging filehandle to only capture errors
    # specific to the particular flatfile.
    if logfile_path is not None:
        flatfile_logger = logging.FileHandler(logfile_path, mode="w")
        flatfile_logger.setLevel(logging.ERROR)
        formatter = logging.Formatter('%(levelname)s: %(message)s')
        flatfile_logger.setFormatter(formatter)
        logger.addHandler(flatfile_logger)

    logger.info("Logging all evaluations.")
    for key1 in dict_of_dict_of_lists:
        msg1 = f"Evaluations for bundle {key1}:"
        logger.info(msg1)
        dict_of_lists = dict_of_dict_of_lists[key1]
        for key2 in dict_of_lists:
            msg2 = f"Evaluations for {key2}:"
            logger.info(msg2)
            evl_list = dict_of_lists[key2]
            for evl in evl_list:
                msg3 = str(evl)
                if evl.status == "warning":
                    logger.warning(msg3)
                elif evl.status == "error":
                    logger.error(msg3)
                else:
                    logger.info(msg3)

    if logfile_path is not None:
        flatfile_logger.close()
        logger.removeHandler(flatfile_logger)


def prepare_tickets(import_table_file=pathlib.Path(), eval_data_dict=None,
        description_field="", table_structure_dict={}):
    """Prepare dictionary of pdm_utils ImportTickets.

    :param import_table_file: same as for data_io().
    :param description_field: same as for data_io().
    :param eval_data_dict:
        Evaluation data dictionary
        Key1 = "eval_mode"
        Value1 = Name of the eval_mode
        Key2 = "eval_flag_dict"
        Value2 = Dictionary of evaluation flags.
    :type eval_data_dict: dict
    :param table_structure_dict:
        Dictionary describing structure of the import table.
    :type table_structure_dict: dict
    :returns:
        Dictionary of pdm_utils ImportTicket objects.
        If a problem was encountered parsing the import table, None is returned.
    :rtype: dict
    """
    # 1. Parse ticket data from table.
    # 2. Set case for certain fields and set default values for missing fields.
    # 3. Add the eval_flag dictionary and description_field to each ticket
    #    provided from command line arguments if they are not provided in the
    #    import table.
    # 4. Check for duplicated phage_ids and ticket ids.
    # 5. For each ticket, evaluate the tickets to ensure they are
    #    structured properly. At this point, the quality of the
    #    ticket data is not evaluated, just that the ticket contains
    #    fields populated or empty as expected.
    # 6. Create a dictionary of tickets based on the phage_id.
    required_keys = table_structure_dict["required"]
    optional_keys = table_structure_dict["optional"]
    keywords = table_structure_dict["keywords"]
    valid_retain = table_structure_dict["valid_retain"]
    valid_retrieve = table_structure_dict["valid_retrieve"]
    valid_add = table_structure_dict["valid_add"]
    valid_parse = table_structure_dict["valid_parse"]

    list_of_tkts = []
    tkt_errors = 0
    logger.info("Retrieving ticket data.")
    list_of_data_dicts = fileio.retrieve_data_dict(import_table_file)
    logger.info("Constructing tickets.")
    list_of_tkts = tickets.construct_tickets(list_of_data_dicts, eval_data_dict,
                    description_field, required_keys, optional_keys,
                    keywords)
    if len(list_of_data_dicts) != len(list_of_tkts):
        tkt_errors += 1

    logger.info("Identifying duplicate tickets.")
    tkt_id_dupes, phage_id_dupes = tickets.identify_duplicates(list_of_tkts)

    ticket_dict = {}
    for x in range(len(list_of_tkts)):
        tkt = list_of_tkts[x]
        tkt_summary = (f"ID: {tkt.id}. "
                       f"Type: {tkt.type}. "
                       f"PhageID: {tkt.phage_id}.")
        logger.info(f"Checking ticket structure for: {tkt_summary}")
        string_list = []
        for key in tkt.eval_flags:
            string_list.append(f"{key}: {tkt.eval_flags[key]}")
        logger.info(f"Evaluation flags are {', '.join(string_list)}")
        check_ticket(tkt,
                     type_set=constants.IMPORT_TICKET_TYPE_SET,
                     description_field_set=constants.DESCRIPTION_FIELD_SET,
                     eval_mode_set=eval_modes.EVAL_MODES.keys(),
                     id_dupe_set=tkt_id_dupes,
                     phage_id_dupe_set=phage_id_dupes,
                     retain_set=valid_retain,
                     retrieve_set=valid_retrieve,
                     add_set=valid_add,
                     parse_set=valid_parse)
        for evl in tkt.evaluations:
            evl_summary = (f"Evaluation: {evl.id}. "
                           f"Status: {evl.status}. "
                           f"Definition: {evl.definition} "
                           f"Result: {evl.result}")
            if evl.status == "error":
                tkt_errors += 1
                logger.error(evl_summary)
            else:
                logger.info(evl_summary)
        ticket_dict[tkt.phage_id] = tkt

    if tkt_errors > 0:
        logger.error("Error generating tickets from import table.")
        return None
    else:
        logger.info("Tickets were successfully generated from import table.")

        return ticket_dict


def process_files_and_tickets(ticket_dict, files_in_folder, engine=None,
                              prod_run=False, genome_id_field="",
                              host_genus_field="", interactive=False,
                              log_folder_paths_dict=None,
                              accept_warning=False):
    """Process GenBank-formatted flat files and import tickets.

    :param ticket_dict:
        A dictionary
        WHERE
        key (str) = The ticket's phage_id
        value (Ticket) = The ticket
    :type ticket_dict: dict
    :param files_in_folder: A list of filepaths to be parsed.
    :type files_in_folder: list
    :param engine: same as for data_io().
    :param prod_run: same as for data_io().
    :param genome_id_field: same as for data_io().
    :param host_genus_field: same as for data_io().
    :param interactive: same as for data_io().
    :param accept_warning: same as for data_io().
    :param log_folder_paths_dict:
        Dictionary indicating paths to success and fail folders.
    :type log_folder_paths_dict: dict
    :returns:
        tuple of five objects
        WHERE
        [0] success_ticket_list (list) is a list of successful ImportTickets.
        [1] failed_ticket_list (list) is a list of failed ImportTickets.
        [2] success_filepath_list (list) is a list of successfully parsed flat files.
        [3] failed_filepath_list (list) is a list of unsuccessfully parsed flat files.
        [4] evaluation_dict (dict):
        dictionary from each Bundle,
        containing dictionaries for each bundled object,
        containing lists of evaluation objects.
    :rtype: tuple
    """
    # Alias for different types of genomes gathered and processed.
    file_ref = "flat_file"
    ticket_ref = "ticket"
    retain_ref = "mysql"
    retrieve_ref = "phagesdb"

    # Will hold all results data. These can be used by various types of
    # user interfaces to present the summary of import.
    success_ticket_list = []
    failed_ticket_list = []
    success_filepath_list = []
    failed_filepath_list = []
    evaluation_dict = {}

    # TODO should add a flag to set default empty values
    # for external data if the import doesn't need to rely on PhagesDB.
    # This would be for building non-Actinobacteriophage databases
    # if retrieve_phagesdb == True:
    #    external_ref_data = get_phagesdb_reference_sets()
    # else:
    #   external_ref_data = <empty data dictionary>

    # Retrieve valid cluster, subcluster, host data from PhagesDB.
    external_ref_data = get_phagesdb_reference_sets()

    # To minimize memory usage, each flat_file is evaluated one by one.
    bundle_count = 1
    file_count = 1
    for filepath in files_in_folder:
        replace_gnm_pair_key = file_ref + "_" + retain_ref
        progress = f"Processing data for file #{file_count}: {filepath.name}."
        print("\n\n" + progress)
        logger.info(progress)

        bndl = prepare_bundle(filepath=filepath, ticket_dict=ticket_dict,
                              engine=engine,
                              genome_id_field=genome_id_field,
                              host_genus_field=host_genus_field,
                              id=bundle_count,
                              file_ref=file_ref, ticket_ref=ticket_ref,
                              retrieve_ref=retrieve_ref, retain_ref=retain_ref,
                              interactive=interactive,
                              id_conversion_dict=constants.PHAGE_ID_DICT)

        # Create sets of unique values for different data fields.
        # Since data from each parsed flat file is imported into the
        # database one file at a time, these sets are not static.
        # So these sets should be recomputed for every flat file evaluated.
        # Retrieve valid data from MySQL and merge with the valid external data.
        mysql_ref_data = get_mysql_reference_sets(engine)
        ref_data = basic.merge_set_dicts(external_ref_data, mysql_ref_data)
        logger.info(f"Checking file: {filepath.name}.")
        run_checks(bndl,
                   accession_set=ref_data["accession_set"],
                   phage_id_set=ref_data["phage_id_set"],
                   seq_set=ref_data["seq_set"],
                   host_genus_set=ref_data["host_genera_set"],
                   cluster_set=ref_data["cluster_set"],
                   subcluster_set=ref_data["subcluster_set"],
                   file_ref=file_ref, ticket_ref=ticket_ref,
                   retrieve_ref=retrieve_ref, retain_ref=retain_ref)

        review_bundled_objects(bndl, interactive=interactive,
                               accept_warning=accept_warning)

        # TODO this section below could probably be improved.
        # import_into_db may not need to return result, since it now
        # records in an Evaluation object whether import succeeded or not.
        # It also checks_for_errors twice (before and after attempting to add
        # data to MySQL), which can probably be simplified.
        bndl.check_for_errors()
        result = import_into_db(bndl, engine=engine,
                                gnm_key=file_ref, prod_run=prod_run)
        bndl.check_for_errors()
        dict_of_eval_lists = bndl.get_evaluations()
        logfile_path = get_logfile_path(bndl, paths_dict=log_folder_paths_dict,
                                        filepath=filepath, file_ref=file_ref)
        log_evaluations({bndl.id: dict_of_eval_lists}, logfile_path=logfile_path)
        evaluation_dict[bndl.id] = dict_of_eval_lists

        if result:
            success_ticket_list.append(bndl.ticket.data_dict)
            success_filepath_list.append(filepath)
        else:
            if bndl.ticket is not None:
                failed_ticket_list.append(bndl.ticket.data_dict)
            failed_filepath_list.append(filepath)
        bundle_count += 1
        file_count += 1

    # Tickets were popped off the ticket dictionary as they were matched
    # to flat files. If there are any tickets left, errors need to be counted.
    if len(ticket_dict.keys()) > 0:
        logger.info("Processing unmatched tickets.")
        key_list = list(ticket_dict.keys())
        for key in key_list:
            bndl = bundle.Bundle()
            bndl.ticket = ticket_dict.pop(key)
            bndl.id = bundle_count
            logger.info("Checking data for ticket: "
                        f"{bndl.ticket.id}, {bndl.ticket.phage_id}.")
            # Since the ticket has no matching genome data,
            # no data sets from MySQL and PhagesDB are needed.
            run_checks(bndl, file_ref=file_ref, ticket_ref=ticket_ref,
                       retrieve_ref=retrieve_ref, retain_ref=retain_ref)
            bndl.check_for_errors()
            dict_of_eval_lists = bndl.get_evaluations()
            logfile_path = get_logfile_path(bndl,
                                            paths_dict=log_folder_paths_dict,
                                            filepath=None, file_ref=None)
            log_evaluations({bndl.id: dict_of_eval_lists}, logfile_path=logfile_path)
            evaluation_dict[bndl.id] = dict_of_eval_lists
            failed_ticket_list.append(bndl.ticket.data_dict)
            bundle_count += 1

    return (success_ticket_list, failed_ticket_list, success_filepath_list,
            failed_filepath_list, evaluation_dict)


def get_phagesdb_reference_sets():
    """Get multiple sets of data from PhagesDB for reference.

    :returns:
        Dictionary of unique clusters, subclusters, and host genera
        stored on PhagesDB.
    :rtype: dict
    """

    # Retrieve data from PhagesDB to create sets of
    # valid host genera, clusters, and subclusters.
    # If there is no subcluster, value may be empty string or "none".
    host_genera = phagesdb.create_host_genus_set()
    results_tuple = phagesdb.create_cluster_subcluster_sets()
    clusters = results_tuple[0]
    subclusters = results_tuple[1]
    dict = {"host_genera_set": host_genera,
            "cluster_set": clusters,
            "subcluster_set": subclusters}
    return dict


def get_mysql_reference_sets(engine):
    """Get multiple sets of data from the MySQL database for reference.

    :param engine: same as for data_io().
    :returns:
        Dictionary of unique PhageIDs, clusters, subclusters,
        host genera, accessions, and sequences stored in the MySQL database.
    :rtype: dict
    """
    phage_ids = mysqldb_basic.get_distinct(engine, "phage", "PhageID")
    accessions = mysqldb_basic.get_distinct(engine, "phage", "Accession")
    clusters = mysqldb_basic.get_distinct(engine, "phage", "Cluster",
                                         null="Singleton")

    # Cluster "UNK" may or may not already be present, but it is valid.
    clusters.add("UNK")
    subclusters = mysqldb_basic.get_distinct(engine, "phage", "Subcluster",
                                            null="none")
    host_genera = mysqldb_basic.get_distinct(engine, "phage", "HostGenus")
    seqs = mysqldb.create_seq_set(engine)
    dict = {"phage_id_set": phage_ids,
            "accession_set": accessions,
            "seq_set": seqs,
            "host_genera_set": host_genera,
            "cluster_set": clusters,
            "subcluster_set": subclusters}
    return dict


def get_logfile_path(bndl, paths_dict=None, filepath=None, file_ref=None):
    """Choose the path to output the file-specific log.

    :param bndl: same as for run_checks().
    :param paths_dict:
        Dictionary indicating paths to success and fail folders.
    :type paths_dict: dict
    :param filepath: Path to flat file.
    :type filepath: Path
    :param file_ref: same as for prepare_bundle().
    :returns:
        Path to log file to store flat-file-specific evaluations.
        If paths_dict is set to None, then None is returned instead of a path.
    :rtype: Path
    """
    # Filename refers to the file that was attempted to be parsed.
    # The genome object contains the filename data only if the file was
    # successfully parsed. If it is not successfully parsed, there would
    # be no genome object to retrieve the filename.
    if paths_dict is not None:
        if bndl._errors > 0:
            log_folder_path = paths_dict[FAIL_FOLDER]
        else:
            log_folder_path = paths_dict[SUCCESS_FOLDER]

        if filepath is None:
            # If there is no filepath, then it is an unmatched ticket.
            id = bndl.ticket.phage_id
            filename = "no_file"
        else:
            # If there is a filepath, get the parsed genome id.
            filename = filepath.stem
            if file_ref in bndl.genome_dict.keys():
                gnm = bndl.genome_dict[file_ref]
                id = gnm.id
            else:
                id = "no_id"
        new_file_name = id + "__" + filename + ".log"
        logfile_path = pathlib.Path(log_folder_path, new_file_name)
    else:
        logfile_path = None
    return logfile_path


def prepare_bundle(filepath=pathlib.Path(), ticket_dict={}, engine=None,
                   genome_id_field="", host_genus_field="", id=None,
                   file_ref="", ticket_ref="", retrieve_ref="", retain_ref="",
                   id_conversion_dict={}, interactive=False):
    """Gather all genomic data needed to evaluate the flat file.

    :param filepath: Name of a GenBank-formatted flat file.
    :type filepath: Path
    :param ticket_dict: A dictionary of pdm_utils ImportTicket objects.
    :type ticket_dict: dict
    :param engine: same as for data_io().
    :param genome_id_field: same as for data_io().
    :param host_genus_field: same as for data_io().
    :param id: Identifier to be assigned to the Bundle object.
    :type id: int
    :param file_ref: Identifier for Genome objects derived from flat files.
    :type file_ref: str
    :param ticket_ref: Identifier for Genome objects derived from ImportTickets.
    :type ticket_ref: str
    :param retrieve_ref: Identifier for Genome objects derived from PhagesDB.
    :type retrieve_ref: str
    :param retain_ref: Identifier for Genome objects derived from MySQL.
    :type retain_ref: str
    :param id_conversion_dict: Dictionary of PhageID conversions.
    :type id_conversion_dict: dict
    :param interactive: same as for data_io().
    :returns:
        A pdm_utils Bundle object containing all data required to
        evaluate a flat file.
    :rtype: Bundle
    """
    bndl = bundle.Bundle()
    bndl.id = id
    seqrecord = flat_files.retrieve_genome_data(filepath)
    if seqrecord is None:
        logger.error(f"No record was retrieved from the file: {filepath.name}.")
    else:
        logger.info(f"Parsing record from the file: {filepath.name}.")
        ff_gnm = flat_files.parse_genome_data(
                    seqrecord,
                    filepath=filepath,
                    genome_id_field=genome_id_field,
                    gnm_type=file_ref,
                    host_genus_field=host_genus_field)

        # Some phage names in the flat file are spelled differently
        # than in the database. In order to match the flat file
        # with the ticket (which must contain the spelling used
        # in the database), the genome ID needs to be changed.
        # id_conversion_dict key = incorrect spelling; value = correct spelling
        if ff_gnm.id in id_conversion_dict.keys():
            ff_gnm.update_name_and_id(id_conversion_dict[ff_gnm.id])

        bndl.genome_dict[ff_gnm.type] = ff_gnm

        # Match ticket (if available) to flat file.
        bndl.ticket = ticket_dict.pop(ff_gnm.id, None)
        if bndl.ticket is None:
            logger.info(f"No matched ticket for file: {filepath.name}.")
        else:
            logger.info(f"Preparing ticket data for file: {filepath.name}.")
            # With the flat file parsed and matched
            # to a ticket, use the ticket to populate specific
            # genome-level fields such as host, cluster, subcluster, etc.
            # Genome attributes from the ticket table that should not be
            # populated from the ticket, from PhagesDB, or from
            # the MySQL database, are stored in data_parse.
            # There is no need to evaluate what is stored in data_parse.

            if len(bndl.ticket.data_add) > 0:
                tkt_gnm = tickets.get_genome(bndl.ticket, gnm_type=ticket_ref)
                bndl.genome_dict[tkt_gnm.type] = tkt_gnm

                # Copy ticket data to flat file. Since the ticket data has been
                # added to a genome object using genome methods, the data
                # can be directly passed from one genome object to another.
                for attr in bndl.ticket.data_add:
                    attr_value = getattr(tkt_gnm, attr)
                    setattr(ff_gnm, attr, attr_value)

            # Check to see if there is any missing data for each genome, and
            # retrieve it from PhagesDB.
            # If the ticket genome has fields set to 'retrieve', data is
            # retrieved from PhagesDB and populates a new Genome object.
            if len(bndl.ticket.data_retrieve) > 0:
                pdb_gnm = phagesdb.get_genome(bndl.ticket.phage_id,
                                              gnm_type=retrieve_ref)
                if pdb_gnm is not None:
                    bndl.genome_dict[pdb_gnm.type] = pdb_gnm

                    for attr in bndl.ticket.data_retrieve:
                        attr_value = getattr(pdb_gnm, attr)
                        setattr(ff_gnm, attr, attr_value)

            # If the ticket type is 'replace', retrieve data from the MySQL database.
            # If any attributes in flat_file are set to 'retain', copy data
            # from the MySQL genome.
            if bndl.ticket.type == "replace":

                if engine is None:
                    logger.info(
                          f"Ticket {bndl.ticket.id} is a 'replace' ticket "
                          "but no details about how to connect to the "
                          "MySQL database have been provided. "
                          "Unable to retrieve data.")
                else:
                    query = "SELECT * FROM phage"
                    pmr_genomes =  mysqldb.parse_genome_data(
                                       engine=engine,
                                       phage_id_list=[ff_gnm.id],
                                       phage_query=query,
                                       gnm_type=retain_ref)
                    if len(pmr_genomes) == 1:
                        pmr_gnm = pmr_genomes[0]
                        bndl.genome_dict[pmr_gnm.type] = pmr_gnm

                        # The ticket may indicate some data should be retained.
                        for attr in bndl.ticket.data_retain:
                            attr_value = getattr(pmr_gnm, attr)
                            setattr(ff_gnm, attr, attr_value)

                        # Pair the genomes for comparison evaluations.
                        gnm_pair = genomepair.GenomePair()
                        bndl.set_genome_pair(gnm_pair, ff_gnm.type, pmr_gnm.type)
                    else:
                        logger.info(f"There is no {ff_gnm.id} genome "
                                    "in the MySQL database. "
                                    "Unable to retrieve data.")

            set_cds_descriptions(ff_gnm, bndl.ticket, interactive=interactive)
    return bndl


def run_checks(bndl, accession_set=set(), phage_id_set=set(),
               seq_set=set(), host_genus_set=set(), cluster_set=set(),
               subcluster_set=set(), file_ref="", ticket_ref="",
               retrieve_ref="", retain_ref=""):
    """Run checks on the different types of data in a Bundle object.

    :param bndl: A pdm_utils Bundle object containing bundled data.
    :type bndl: Bundle
    :param accession_set: Set of accessions to check against.
    :type accession_set: set
    :param phage_id_set: Set of PhageIDs to check against.
    :type phage_id_set: set
    :param seq_set: Set of nucleotide sequences to check against.
    :type seq_set: set
    :param host_genus_set: Set of host genera to check against.
    :type host_genus_set: set
    :param cluster_set: Set of Clusters to check against.
    :type cluster_set: set
    :param subcluster_set: Set of Subclusters to check against.
    :type subcluster_set: set
    :param file_ref: same as for prepare_bundle().
    :param ticket_ref: same as for prepare_bundle().
    :param retrieve_ref: same as for prepare_bundle().
    :param retain_ref: same as for prepare_bundle().
    """
    logger.info("Checking data.")
    check_bundle(bndl, ticket_ref=ticket_ref, file_ref=file_ref,
                 retrieve_ref=retrieve_ref, retain_ref=retain_ref)
    tkt = bndl.ticket
    if tkt is not None:
        eval_flags = tkt.eval_flags
        gnm_pair_key = file_ref + "_" + retain_ref
        if (tkt.type == "replace" and
                gnm_pair_key in bndl.genome_pair_dict.keys()):
            genome_pair = bndl.genome_pair_dict[gnm_pair_key]
            compare_genomes(genome_pair, eval_flags)

        if file_ref in bndl.genome_dict.keys():
            gnm = bndl.genome_dict[file_ref]
            check_genome(gnm, tkt.type, eval_flags,
                         accession_set=accession_set, phage_id_set=phage_id_set,
                         seq_set=seq_set, host_genus_set=host_genus_set,
                         cluster_set=cluster_set, subcluster_set=subcluster_set)

            # Check each type of feature.
            for x in range(len(gnm.cds_features)):
                check_cds(gnm.cds_features[x], eval_flags,
                          description_field=tkt.description_field)

            for x in range(len(gnm.trna_features)):
                check_trna(gnm.trna_features[x], eval_flags)

            for x in range(len(gnm.tmrna_features)):
                check_tmrna(gnm.tmrna_features[x], eval_flags)

            for x in range(len(gnm.source_features)):
                check_source(gnm.source_features[x], eval_flags,
                             host_genus=gnm.host_genus)

        if retain_ref in bndl.genome_dict.keys():
            gnm2 = bndl.genome_dict[retain_ref]
            check_retain_genome(gnm2, tkt.type, eval_flags)


def review_bundled_objects(bndl, interactive=False, accept_warning=False):
    """Review all evaluations of all bundled objects.

    Iterate through all objects stored in the bundle.
    If there are warnings, review whether status should be changed.

    :param bndl: same as for run_checks().
    :param interactive: same as for data_io().
    :param accept_warnring: same as for data_io().
    """

    # Bundle-level evaluations.
    review_object_list([bndl], "Bundle", ["id"], interactive=interactive,
                       accept_warning=accept_warning)

    # Ticket-level evaluations.
    review_object_list([bndl.ticket], "Ticket", ["id", "type", "phage_id"],
                       interactive=interactive, accept_warning=accept_warning)

    # Genome check.
    if len(bndl.genome_dict.keys()) > 0:
        for key in bndl.genome_dict.keys():
            gnm = bndl.genome_dict[key]

            # Genome-level evaluations.
            # Instead of passing all genomes from the genome_dict at one time,
            # it makes more sense to evaluate only one genome.evaluations list
            # at the same time as the evaluations for its cds, tRNA, tmRNA,
            # and source features.
            review_object_list([gnm], "Genome", ["id", "type"],
                               interactive=interactive,
                               accept_warning=accept_warning)

            review_object_list(gnm.cds_features, "CDS feature",
                               ["id", "start", "stop", "orientation"],
                               interactive=interactive,
                               accept_warning=accept_warning)

            review_object_list(gnm.source_features, "Source feature",
                               ["id"], interactive=interactive,
                               accept_warning=accept_warning)

            review_object_list(gnm.trna_features, "tRNA feature",
                               ["id", "start", "stop", "orientation"],
                               interactive=interactive,
                               accept_warning=accept_warning)

            review_object_list(gnm.tmrna_features, "tmRNA feature",
                               ["id", "start", "stop", "orientation"],
                               interactive=interactive,
                               accept_warning=accept_warning)

    else:
        log_and_print("No genomes to review.", False)

    # Genome-pair check.
    genome_pair_list = list(bndl.genome_pair_dict.values())
    review_object_list(genome_pair_list, "Genome pair",
                       ["genome1","genome2"], interactive=interactive,
                       accept_warning=accept_warning)


def review_object_list(object_list, object_type, attr_list, interactive=False,
                       accept_warning=False):
    """Determine if evaluations are present and record results.

    :param object_list: List of pdm_utils objects containing evaluations.
    :type object_list: list
    :param object_type: Name of the pdm_utils object.
    :type object_type: str
    :param attr_list:
        List of attributes used to log data about the object instance.
    :type attr_list: list
    :param interactive: same as for data_io().
    :param accept_warning: same as for data_io().
    """
    # Test for None, since tkt data can be missing and be None.
    if (len(object_list) > 0 and object_list[0] is not None):
        # Capture the exit status for each CDS feature. If user exits
        # the review at any point, skip all of the other CDS features.
        exit = False
        for x in range(len(object_list)):
            object = object_list[x]
            # Compile description of the object being reviewed.
            string = get_result_string(object, attr_list)
            partial_msg = f"evaluations for {object_type}: {string}."
            if len(object.evaluations) > 0:
                log_and_print("Reviewing " + partial_msg, interactive)
                if exit == False:
                    exit = review_evaluation_list(
                                                object.evaluations,
                                                interactive=interactive,
                                                accept_warning=accept_warning)
                else:
                    # If exit=True, all 'warning' evaluations are automatically
                    # changed to 'error'. The exit response is not captured.
                    review_evaluation_list(object.evaluations,
                                           interactive=False,
                                           accept_warning=False)
            else:
                log_and_print("No " + partial_msg, False)
        if exit:
            msg = ("Not all evaluations were reviewed, "
                   "since the review was exited.")
            log_and_print(msg, False)
    else:
        log_and_print(f"No evaluations for {object_type}(s)", False)


def review_evaluation_list(evaluation_list, interactive=False,
                           accept_warning=False):
    """Iterate through all evaluations and review 'warning' results.

    :param evaluation_list: List of pdm_utils Evaluation objects.
    :type evaluation_list: list
    :param interactive: same as for data_io().
    :param accept_warning: same as for data_io().
    :returns: Indicates whether user selected to exit the review process.
    :rtype: bool
    """
    exit = False
    y = 0
    for x in range(len(evaluation_list)):
        evl = evaluation_list[x]
        if exit == False:
            exit, correct = review_evaluation(evl, interactive=interactive,
                                              accept_warning=accept_warning)
        else:
            # If exit=True, then all 'warning' evaluations are automatically
            # changed to 'error'. The exit2 response is unused and thrown away.
            exit2, correct = review_evaluation(evl, interactive=False)
        if not correct:
            y += 1

    if y == 0:
        log_and_print(f"No warnings or errors encountered.", interactive)
    return exit


def review_evaluation(evl, interactive=False, accept_warning=False):
    """Review an evaluation object.

    :param evl: A pdm_utils Evaluation object.
    :type evl: Evaluation
    :param interactive: same as for data_io().
    :param accept_warning: same as for data_io().
    :returns:
        tuple (exit, message)
        WHERE
        exit (bool) indicates whether user selected to exit the review process.
        correct(bool) indicates whether the evalution status is accurate.
    :rtype: tuple
    """
    exit = False
    msg = " Status was changed from 'warning' to 'error' {}."
    summary = (f"Evaluation ID: {evl.id}."
               f"\nStatus: {evl.status}."
               f"\nDefinition: {evl.definition}"
               f"\nResult: {evl.result}")
    if evl.status == "warning":
        correct = False
        if interactive == True:
            # If interactive is set to True, ask user if 'warning'
            # is correct, and change the status as needed.
            print("\n\nThe following evaluation is set to 'warning':")
            print(summary)
            prompt = ("\nThis evaluation will remain as a 'warning' "
                      "(instead of it being changed to an 'error'). "
                      "Is this correct? (yes/no/exit) ")
            result = basic.ask_yes_no(prompt=prompt, response_attempt=3)
            # result can be True, False, or None
            if result != True:
                evl.status = "error"
                if result == False:
                    evl.result = evl.result + msg.format("manually")
                else:
                    exit = True
                    evl.result = evl.result + msg.format(
                                    "automatically due to review exit")
        elif accept_warning:
            pass
        else:
            # If interactive is set to False,
            # change all 'warnings' to 'errors'.
            evl.status = "error"
            evl.result = evl.result + msg.format(
                            "automatically due to no interactivity")
    elif evl.status == "error":
        correct = False
        if interactive == True:
            # Report that an error was encountered.
            print("\n\nThe following evaluation is set to 'error':")
            print(summary)
            input("\nPress ENTER to continue.")
    else:
        correct = True
    return exit, correct


def log_and_print(msg, terminal=False):
    """Print message to terminal in addition to logger if needed.

    :param msg: Message to print.
    :type msg: str
    :param terminal: Indicates if message should be printed to terminal.
    :type terminal: bool
    """
    logger.info(msg)
    if terminal:
        print(f"{msg}")


def get_result_string(object, attr_list):
    """Construct string of values from several object attributes.

    :param object: A object from which to retrieve values.
    :type object: misc
    :param attr_list:
        List of strings indicating attributes to retrieve from the object.
    :type attr_list: list
    :returns: A concatenated string representing values from all attributes.
    :rtype: str
    """
    strings = []
    for attr in attr_list:
        attr_value = getattr(object, attr)
        strings.append(attr + ": " + str(attr_value))
    string = ", ".join(strings)
    return string


def set_cds_descriptions(gnm, tkt, interactive=False):
    """Set the primary CDS descriptions.

    :param gnm: A pdm_utils Genome object.
    :type gnm: Genome
    :param tkt: A pdm_utils ImportTicket object.
    :type tkt: ImportTicket
    :param interactive: same as for data_io().
    """
    # If interactive is selected, the user can review the CDS descriptions.
    # The ticket indicates where the CDS descriptions are stored.
    if interactive:
        tkt.description_field = review_cds_descriptions(
                                    gnm.cds_features,
                                    tkt.description_field)
    gnm.set_cds_descriptions(tkt.description_field)
    gnm.tally_cds_descriptions()


def review_cds_descriptions(feature_list, description_field):
    """Iterate through all CDS features and review descriptions.

    :param feature_list: A list of pdm_utils Cds objects.
    :type feature_list: list
    :param description_field: same as for data_io().
    :returns: Name of the primary description_field after review.
    :rtype: str
    """
    logger.info("Reviewing CDS description fields.")

    # Print a table of CDS data for user to review.
    summary = []
    x = 20
    y = "..."
    tally = 0
    for cds in feature_list:
        # Get truncated descriptions to report to user.
        short_product = basic.truncate_value(cds.product, x, y)
        short_function = basic.truncate_value(cds.function, x, y)
        short_note = basic.truncate_value(cds.note, x, y)
        dict = {"CDS ID": cds.id,
                "Start": cds.start,
                "Stop": cds.stop,
                "Orientation": cds.orientation,
                "Product": short_product,
                "Function": short_function,
                "Note": short_note
                }
        summary.append(dict)

        # Count the number of descriptions that are not
        # in the selected description field.
        if (description_field != "product" and len(short_product) > 0):
            tally += 1
        if (description_field != "function" and len(short_function) > 0):
            tally += 1
        if (description_field != "note" and len(short_note) > 0):
            tally += 1


    # If there are > 0 descriptions present in unselected description fields,
    # print results and confirm with user.
    result = True
    if tally > 0:
        print(tabulate(summary, headers="keys"))
        print(f"Descriptions in the {description_field} qualifier "
              f"will be imported, but there are {tally} description(s) "
              "detected in other qualifiers.")
        prompt = "Is this correct? "
        result = basic.ask_yes_no(prompt=prompt, response_attempt=3)

    # If the data is not correct, let the user select which description
    # field is most appropriate.
    if result == False:
        field_options = constants.DESCRIPTION_FIELD_SET - {description_field}
        print("Select the field in which the descriptions are stored.")
        new_field = basic.choose_from_list(list(field_options))
        if new_field is None:
            new_field = description_field
            msg = ("No new CDS description field was selected. "
                  f"The '{new_field}' CDS description field will be used.")
        else:
            msg = (f"The '{new_field}' CDS description field was selected.")

    else:
        new_field = description_field
        msg = (f"The '{new_field}' CDS description field is correct, "
               "so it will be used.")
    print(msg)
    logger.info(msg)

    # Return either the original or the new description field.
    return new_field


def check_bundle(bndl, ticket_ref="", file_ref="", retrieve_ref="",
                 retain_ref=""):
    """Check a Bundle for errors.

    Evaluate whether all genomes have been successfully grouped,
    and whether all genomes have been paired, as expected.
    Based on the ticket type, there are expected to be certain
    types of genomes and pairs of genomes in the bundle.

    :param bndl: same as for run_checks().
    :param ticket_ref: same as for prepare_bundle().
    :param file_ref: same as for prepare_bundle().
    :param retrieve_ref: same as for prepare_bundle().
    :param retain_ref: same as for prepare_bundle().
    """
    logger.info(f"Checking bundle: {bndl.id}.")

    # Log file of genome, if applicable.
    if file_ref in bndl.genome_dict.keys():
        gnm = bndl.genome_dict[file_ref]
        logger.info(f"Genome from file: {gnm.filename}.")

    bndl.check_ticket(eval_id="BNDL-EVAL-001", eval_def=EDD["BNDL-EVAL-001"])
    if bndl.ticket is not None:
        logger.info(f"Ticket: {bndl.ticket.type}, {bndl.ticket.phage_id}.")
        bndl.check_genome_dict(file_ref, expect=True, eval_id="BNDL-EVAL-002",
                               eval_def=EDD["BNDL-EVAL-002"])
        tkt = bndl.ticket
        if len(tkt.data_add) > 0:
            bndl.check_genome_dict(ticket_ref, expect=True,
                                   eval_id="BNDL-EVAL-003",
                                   eval_def=EDD["BNDL-EVAL-003"])
        # There may or may not be data retrieved from PhagesDB.
        if len(tkt.data_retrieve) > 0:
            bndl.check_genome_dict(retrieve_ref,
                                   expect=True, eval_id="BNDL-EVAL-004",
                                   eval_def=EDD["BNDL-EVAL-004"])
        if tkt.type == "replace":
            bndl.check_genome_dict(retain_ref, expect=True,
                                   eval_id="BNDL-EVAL-005",
                                   eval_def=EDD["BNDL-EVAL-005"])
            # There should be a genome_pair between the current MySQL
            # genome and the new flat_file genome.
            pair_key = f"{file_ref}_{retain_ref}"
            bndl.check_genome_pair_dict(pair_key, eval_id="BNDL-EVAL-006",
                                        eval_def=EDD["BNDL-EVAL-006"])


def check_ticket(tkt, type_set=set(), description_field_set=set(),
        eval_mode_set=set(), id_dupe_set=set(), phage_id_dupe_set=set(),
        retain_set=set(), retrieve_set=set(), add_set=set(), parse_set=set()):
    """Evaluate a ticket to confirm it is structured appropriately.

    The assumptions for how each field is populated varies depending on
    the type of ticket.

    :param tkt: same as for set_cds_descriptions().
    :param type_set: Set of ImportTicket types to check against.
    :type type_set: set
    :param description_field_set: Set of description fields to check against.
    :type description_field_set: set
    :param eval_mode_set: Set of evaluation modes to check against.
    :type eval_mode_set: set
    :param id_dupe_set: Set of duplicated ImportTicket ids to check against.
    :type id_dupe_set: set
    :param phage_id_dupe_set: Set of duplicated ImportTicket PhageIDs to check against.
    :type phage_id_dupe_set: set
    :param retain_set: Set of retain values to check against.
    :type retain_set: set
    :param retrieve_set: Set of retrieve values to check against.
    :type retrieve_set: set
    :param add_set: Set of add values to check against.
    :type add_set: set
    :param parse_set: Set of parse values to check against.
    :type parse_set: set
    """
    # This function simply evaluates whether there is data in the
    # appropriate ticket attributes given the type of ticket.
    # It confirms that ticket attributes 'type', 'eval_mode', and
    # 'description_field' are populated with specific values.
    # But it does not evaluate the quality of the data itself for
    # the other fields, since those are genome-specific fields and
    # can be checked within Genome objects.
    # logger.info(f"Checking ticket: {tkt.id}, {tkt.type}, {tkt.phage_id}.")

    # Check for duplicated values.
    tkt.check_attribute("id", id_dupe_set,
                        expect=False, eval_id="TKT-EVAL-001",
                        eval_def=EDD["TKT-EVAL-001"])
    tkt.check_attribute("phage_id", phage_id_dupe_set,
                        expect=False, eval_id="TKT-EVAL-002",
                        eval_def=EDD["TKT-EVAL-002"])

    # Check these fields for specific values.
    tkt.check_attribute("type", type_set,
                        expect=True, eval_id="TKT-EVAL-003",
                        eval_def=EDD["TKT-EVAL-003"])
    tkt.check_attribute("description_field", description_field_set,
                        expect=True, eval_id="TKT-EVAL-004",
                        eval_def=EDD["TKT-EVAL-004"])
    tkt.check_attribute("eval_mode", eval_mode_set,
                        expect=True, eval_id="TKT-EVAL-005",
                        eval_def=EDD["TKT-EVAL-005"])

    # This method may be refactored so that it accepts a list of
    # valid flag dict keys. But this has already been verified earlier
    # in the script, so it could be redundant.
    tkt.check_eval_flags(expect=True, eval_id="TKT-EVAL-006",
                         eval_def=EDD["TKT-EVAL-006"])

    # For these fields, simply check that they are not empty.
    tkt.check_attribute("phage_id", {""}, expect=False, eval_id="TKT-EVAL-007",
                        eval_def=EDD["TKT-EVAL-007"])

    # Check how genome attributes will be determined.
    tkt.check_compatible_type_and_data_retain(eval_id="TKT-EVAL-008",
                                              eval_def=EDD["TKT-EVAL-008"])
    tkt.check_valid_data_source("data_add", add_set,
                                eval_id="TKT-EVAL-009",
                                eval_def=EDD["TKT-EVAL-009"])
    tkt.check_valid_data_source("data_retain", retain_set,
                                eval_id="TKT-EVAL-010",
                                eval_def=EDD["TKT-EVAL-010"])
    tkt.check_valid_data_source("data_retrieve", retrieve_set,
                                eval_id="TKT-EVAL-011",
                                eval_def=EDD["TKT-EVAL-011"])
    tkt.check_valid_data_source("data_parse", parse_set,
                                eval_id="TKT-EVAL-012",
                                eval_def=EDD["TKT-EVAL-012"])


def check_genome(gnm, tkt_type, eval_flags, phage_id_set=set(),
                 seq_set=set(), host_genus_set=set(),
                 cluster_set=set(), subcluster_set=set(),
                 accession_set=set()):
    """Check a Genome object parsed from file for errors.

    :param gnm: A pdm_utils Genome object.
    :type gnm: Genome
    :param tkt: A pdm_utils Ticket object.
    :type tkt: Ticket
    :param eval_flags: Dictionary of boolean evaluation flags.
    :type eval_flags: dicts
    :param phage_id_set: Set of PhageIDs to check against.
    :type phage_id_set: set
    :param seq_set: Set of genome sequences to check against.
    :type seq_set: set
    :param host_genus_set: Set of host genera to check against.
    :type host_genus_set: set
    :param cluster_set: Set of clusters to check against.
    :type cluster_set: set
    :param subcluster_set: Set of subclusters to check against.
    :type subcluster_set: set
    :param accession_set: Set of accessions to check against.
    :type accession_set: set
    """
    logger.info(f"Checking genome: {gnm.id}, {gnm.type}.")

    if tkt_type == "add":
        gnm.check_attribute("id", phage_id_set | {""}, expect=False,
                            eval_id="GNM-EVAL-001",
                            eval_def=EDD["GNM-EVAL-001"])
        gnm.check_attribute("name", phage_id_set | {""}, expect=False,
                            eval_id="GNM-EVAL-002",
                            eval_def=EDD["GNM-EVAL-002"])
        gnm.check_attribute("seq", seq_set | {constants.EMPTY_GENOME_SEQ},
                            expect=False, eval_id="GNM-EVAL-003",
                            eval_def=EDD["GNM-EVAL-003"])
        gnm.check_attribute("annotation_status", {"final"}, expect=False,
                            eval_id="GNM-EVAL-004", fail="warning",
                            eval_def=EDD["GNM-EVAL-004"])



        # If the genome is being added, and if it has an accession,
        # no other genome is expected to have an identical accession.
        # If the genome is being replaced, and if it has an accession,
        # the prior version of the genome may or may not have had
        # accession data, so no need to check for 'replace' tickets.
        # Direct comparison of the accession from the old and new genomes
        # occurs in compare_genomes().
        if gnm.accession != "":
            gnm.check_attribute("accession", accession_set, expect=False,
                                eval_id="GNM-EVAL-005",
                                eval_def=EDD["GNM-EVAL-005"])

    # 'replace' ticket checks.
    else:
        gnm.check_attribute("id", phage_id_set, expect=True,
                            eval_id="GNM-EVAL-006",
                            eval_def=EDD["GNM-EVAL-006"])
        gnm.check_attribute("seq", seq_set, expect=True,
                            eval_id="GNM-EVAL-007",
                            eval_def=EDD["GNM-EVAL-007"])
        gnm.check_attribute("annotation_status", {"draft"}, expect=False,
                            eval_id="GNM-EVAL-008", fail="warning",
                            eval_def=EDD["GNM-EVAL-008"])


    # Depending on the annotation_status of the genome,
    # CDS features are expected to contain or not contain descriptions.
    # Draft genomes should not have any descriptions.
    # Final genomes should have any descriptions.
    # There are no expectations for other types of genomes.

    # Also, Draft annotations should not have accession data.

    check_name = basic.edit_suffix(gnm.name, "add",
                                   suffix=constants.NAME_SUFFIX)

    if gnm.annotation_status == "draft":
        gnm.check_attribute("name", {check_name}, expect=True,
                            eval_id="GNM-EVAL-009", fail="warning",
                            eval_def=EDD["GNM-EVAL-009"])
        gnm.check_magnitude("_cds_descriptions_tally", "=", 0,
                            eval_id="GNM-EVAL-010", fail="warning",
                            eval_def=EDD["GNM-EVAL-010"])
        gnm.check_attribute("accession", {""}, expect=True,
                            eval_id="GNM-EVAL-011", fail="warning",
                            eval_def=EDD["GNM-EVAL-011"])

    else:
        if gnm.annotation_status == "final":
            gnm.check_attribute("name", {check_name}, expect=False,
                                eval_id="GNM-EVAL-012", fail="warning",
                                eval_def=EDD["GNM-EVAL-012"])

            if eval_flags["check_description_tally"]:
                gnm.check_magnitude("_cds_descriptions_tally", ">", 0,
                                    eval_id="GNM-EVAL-013", fail="warning",
                                    eval_def=EDD["GNM-EVAL-013"])

    if eval_flags["check_coords"]:
        gnm.check_feature_coordinates(
                    use_cds=True, use_trna=True, use_tmrna=True,
                    strand=False, eval_id="GNM-EVAL-030",
                    eval_def=EDD["GNM-EVAL-030"], fail="warning")

    check_id = basic.edit_suffix(gnm.name, "add", suffix=constants.NAME_SUFFIX)
    gnm.check_attribute("id", {check_id}, expect=False,
                        eval_id="GNM-EVAL-014", fail="warning",
                        eval_def=EDD["GNM-EVAL-014"])
    gnm.check_attribute("annotation_status", constants.ANNOTATION_STATUS_SET,
                        expect=True, eval_id="GNM-EVAL-015",
                        eval_def=EDD["GNM-EVAL-015"])
    gnm.check_attribute("annotation_author", constants.ANNOTATION_AUTHOR_SET,
                        expect=True, eval_id="GNM-EVAL-016",
                        eval_def=EDD["GNM-EVAL-016"])
    gnm.check_attribute("retrieve_record", constants.RETRIEVE_RECORD_SET,
                        expect=True, eval_id="GNM-EVAL-017",
                        eval_def=EDD["GNM-EVAL-017"])
    gnm.check_attribute("cluster", cluster_set, expect=True,
                        eval_id="GNM-EVAL-018", fail="warning",
                        eval_def=EDD["GNM-EVAL-018"])
    gnm.check_attribute("subcluster", subcluster_set | {"none"}, expect=True,
                        eval_id="GNM-EVAL-019", fail="warning",
                        eval_def=EDD["GNM-EVAL-019"])
    gnm.check_attribute("translation_table", {11}, expect=True,
                        eval_id="GNM-EVAL-020", fail="warning",
                        eval_def=EDD["GNM-EVAL-020"])
    gnm.check_attribute("host_genus", host_genus_set, expect=True,
                        eval_id="GNM-EVAL-021", fail="warning",
                        eval_def=EDD["GNM-EVAL-021"])

    gnm.check_cluster_structure(eval_id="GNM-EVAL-022", fail="warning",
                                eval_def=EDD["GNM-EVAL-022"])
    gnm.check_subcluster_structure(eval_id="GNM-EVAL-023", fail="warning",
                                   eval_def=EDD["GNM-EVAL-023"])
    gnm.check_compatible_cluster_and_subcluster(eval_id="GNM-EVAL-024",
                                                eval_def=EDD["GNM-EVAL-024"])
    gnm.check_magnitude("date", ">", constants.EMPTY_DATE, eval_id="GNM-EVAL-025",
                        eval_def=EDD["GNM-EVAL-025"])
    gnm.check_magnitude("gc", ">", -0.0001, eval_id="GNM-EVAL-026",
                        eval_def=EDD["GNM-EVAL-026"])
    gnm.check_magnitude("gc", "<", 100.0001, eval_id="GNM-EVAL-027",
                        eval_def=EDD["GNM-EVAL-027"])
    gnm.check_magnitude("length", ">", 0, eval_id="GNM-EVAL-028",
                        eval_def=EDD["GNM-EVAL-028"])
    gnm.check_magnitude("_cds_features_tally", ">", 0, eval_id="GNM-EVAL-029",
                        fail="warning", eval_def=EDD["GNM-EVAL-029"])

    if eval_flags["check_seq"]:
        gnm.check_nucleotides(check_set=constants.DNA_ALPHABET,
                              eval_id="GNM-EVAL-031", fail="warning",
                              eval_def=EDD["GNM-EVAL-031"])

    if eval_flags["check_id_typo"]:
        gnm.compare_two_attributes("id", "_description_name", expect_same=True,
                                   eval_id="GNM-EVAL-032", fail="warning",
                                   eval_def=EDD["GNM-EVAL-032"])
        gnm.compare_two_attributes("id", "_source_name", expect_same=True,
                                   eval_id="GNM-EVAL-033", fail="warning",
                                   eval_def=EDD["GNM-EVAL-033"])
        gnm.compare_two_attributes("id", "_organism_name", expect_same=True,
                                   eval_id="GNM-EVAL-034", fail="warning",
                                   eval_def=EDD["GNM-EVAL-034"])

    if eval_flags["check_host_typo"]:

        # TODO change these checks in parallel with source feature checks?
        # Could get a set of host genus synonyms and use check_attribute()
        # Actually, it may be better to make source feature checks parallel to
        # these genome checks. May no longer need host genus synonyms.
        # Host genus can be defined by ticket, which is set to gnm.host_genus.
        gnm.compare_two_attributes("host_genus", "_description_host_genus",
                                   expect_same=True, eval_id="GNM-EVAL-035",
                                   fail="warning", eval_def=EDD["GNM-EVAL-035"])
        gnm.compare_two_attributes("host_genus", "_source_host_genus",
                                   expect_same=True, eval_id="GNM-EVAL-036",
                                   fail="warning", eval_def=EDD["GNM-EVAL-036"])
        gnm.compare_two_attributes("host_genus", "_organism_host_genus",
                                   expect_same=True, eval_id="GNM-EVAL-037",
                                   fail="warning", eval_def=EDD["GNM-EVAL-037"])

    if eval_flags["check_author"]:
        if gnm.annotation_author == 1:
            gnm.check_authors(check_set=constants.AUTHOR_SET,
                              expect=True, eval_id="GNM-EVAL-038",
                              fail="warning", eval_def=EDD["GNM-EVAL-038"])
            gnm.check_authors(check_set=set(["lastname", "firstname"]),
                              expect=False, eval_id="GNM-EVAL-039",
                              fail="warning", eval_def=EDD["GNM-EVAL-039"])
        else:
            gnm.check_authors(check_set=constants.AUTHOR_SET,
                              expect=False, eval_id="GNM-EVAL-040",
                              fail="warning", eval_def=EDD["GNM-EVAL-040"])


def check_retain_genome(gnm, tkt_type, eval_flags):
    """Check a Genome object currently in database for errors.

    :param gnm: A pdm_utils Genome object.
    :type gnm: Genome
    :param tkt_type: ImportTicket type
    :type tkt_type: str
    :param eval_flags: Dictionary of boolean evaluation flags.
    :type eval_flags: dicts
    """
    logger.info(f"Checking genome: {gnm.id}, {gnm.type}.")
    if eval_flags["check_replace"]:
        gnm.check_attribute("annotation_status", {"draft"},
                            expect=True, eval_id="GNM2-EVAL-001",
                            fail="warning", eval_def=EDD["GNM2-EVAL-001"])


def check_source(src_ftr, eval_flags, host_genus=""):
    """Check a Source object for errors.

    :param src_ftr: A pdm_utils Source object.
    :type src_ftr: Source
    :param eval_flags: Dictionary of boolean evaluation flags.
    :type eval_flags: dicts
    :param host_genus: Host genus to check against.
    :type host_genus: str
    """
    logger.info(f"Checking source feature: {src_ftr.id}.")

    if eval_flags["check_id_typo"]:
        src_ftr.check_attribute("_organism_name", {src_ftr.genome_id},
                                expect=True, eval_id="SRC-EVAL-001",
                                fail="warning", eval_def=EDD["SRC-EVAL-001"])

    if eval_flags["check_host_typo"]:

        # Alternatively, a host_genus_set could be passed into the function,
        # instead of a string, and the decision to pass in a set of synonyms
        # could be pushed up a level.

        # Only evaluate attributes if the field is not empty, since they
        # are not required to be present.
        if src_ftr.organism != "":
            src_ftr.check_attribute("_organism_host_genus", {host_genus},
                                    expect=True, eval_id="SRC-EVAL-002",
                                    fail="warning", eval_def=EDD["SRC-EVAL-002"])
        if src_ftr.host != "":
            src_ftr.check_attribute("_host_host_genus", {host_genus},
                                    expect=True, eval_id="SRC-EVAL-003",
                                    fail="warning", eval_def=EDD["SRC-EVAL-003"])
        if src_ftr.lab_host != "":
            src_ftr.check_attribute("_lab_host_host_genus", {host_genus},
                                    expect=True, eval_id="SRC-EVAL-004",
                                    fail="warning", eval_def=EDD["SRC-EVAL-004"])


def check_cds(cds_ftr, eval_flags, description_field="product"):
    """Check a Cds object for errors.

    :param cds_ftr: A pdm_utils Cds object.
    :type cds_ftr: Cds
    :param eval_flags: Dictionary of boolean evaluation flags.
    :type eval_flags: dicts
    :param description_field: Description field to check against.
    :type description_field: str
    """
    logger.info(f"Checking CDS feature: {cds_ftr.locus_tag} ({cds_ftr.id}).")

    cds_ftr.check_attribute("translation", {constants.EMPTY_PROTEIN_SEQ},
                            expect=False, eval_id="CDS-EVAL-003",
                            eval_def=EDD["CDS-EVAL-003"])
    cds_ftr.check_amino_acids(check_set=constants.PROTEIN_ALPHABET,
                              fail="warning", eval_id="CDS-EVAL-001",
                              eval_def=EDD["CDS-EVAL-001"])
    cds_ftr.check_translation(eval_id="CDS-EVAL-002",
                              eval_def=EDD["CDS-EVAL-002"])
    cds_ftr.check_attribute("translation_table", {11},
                            expect=True, eval_id="CDS-EVAL-004", fail="warning",
                            eval_def=EDD["CDS-EVAL-004"])
    # Start can be 0 since coordinate_format = 0_half_open
    cds_ftr.check_magnitude("start", ">", -1, eval_id="CDS-EVAL-005",
                            eval_def=EDD["CDS-EVAL-005"])
    cds_ftr.check_magnitude("stop", ">", 0, eval_id="CDS-EVAL-013",
                            eval_def=EDD["CDS-EVAL-013"])
    cds_ftr.check_magnitude("parts", ">", 0, eval_id="CDS-EVAL-014",
                            eval_def=EDD["CDS-EVAL-014"])
    cds_ftr.check_orientation(format="fr_short", case=True,
                              eval_id="CDS-EVAL-006",
                              eval_def=EDD["CDS-EVAL-006"])

    if eval_flags["check_locus_tag"]:
        cds_ftr.check_attribute("locus_tag", {""}, expect=False,
                                eval_id="CDS-EVAL-007", fail="warning",
                                eval_def=EDD["CDS-EVAL-007"])

        # TODO this check could be improved to take into account the prefix.
        cds_ftr.check_locus_tag_structure(check_value=None, only_typo=True,
                                case=True, eval_id="CDS-EVAL-008",
                                fail="warning", eval_def=EDD["CDS-EVAL-008"])

    if eval_flags["check_gene"]:
        cds_ftr.check_attribute("gene", {""}, expect=False,
                                eval_id="CDS-EVAL-009", fail="warning",
                                eval_def=EDD["CDS-EVAL-009"])
        cds_ftr.check_gene_structure(eval_id="CDS-EVAL-010", fail="warning",
                                     eval_def=EDD["CDS-EVAL-010"])

    if (eval_flags["check_locus_tag"] and eval_flags["check_gene"]):
        cds_ftr.check_compatible_gene_and_locus_tag(
                    eval_id="CDS-EVAL-011", fail="warning",
                    eval_def=EDD["CDS-EVAL-011"])

    # if eval_flags["check_description"]:
        # TODO the "check_generic_data" method should be implemented
        # at the genome level.
        # cds_ftr.check_generic_data(eval_id="CDS-EVAL-012")
        # TODO not implemented yet:
        #    cds_ftr.check_valid_description(eval_id="CDS-EVAL-013")
    if eval_flags["check_description_field"]:
        cds_ftr.check_description_field(attribute=description_field,
                                        eval_id="CDS-EVAL-012", fail="warning",
                                        eval_def=EDD["CDS-EVAL-012"])


def compare_genomes(genome_pair, eval_flags):
    """Compare two genomes to identify discrepancies.

    :param genome_pair: A pdm_utils GenomePair object.
    :type genome_pair: GenomePair
    :param eval_flags: Dictionary of boolean evaluation flags.
    :type eval_flags: dicts
    """
    logger.info("Comparing data between two genomes: "
                f"Genome 1. ID: {genome_pair.genome1.id}. "
                f"Type: {genome_pair.genome1.type}. "
                f"Genome 2. ID: {genome_pair.genome2.id}. "
                f"Type: {genome_pair.genome2.type}."
                )

    genome_pair.compare_attribute("id", expect_same=True, eval_id="GP-EVAL-001",
                                  fail="warning", eval_def=EDD["GP-EVAL-001"])
    genome_pair.compare_attribute("seq", expect_same=True, eval_id="GP-EVAL-002",
                                  fail="warning", eval_def=EDD["GP-EVAL-002"])
    genome_pair.compare_attribute("length", expect_same=True,
                                  eval_id="GP-EVAL-003", fail="warning",
                                  eval_def=EDD["GP-EVAL-003"])
    genome_pair.compare_attribute("cluster", expect_same=True,
                                  eval_id="GP-EVAL-004", fail="warning",
                                  eval_def=EDD["GP-EVAL-004"])
    genome_pair.compare_attribute("subcluster", expect_same=True,
                                  eval_id="GP-EVAL-005", fail="warning",
                                  eval_def=EDD["GP-EVAL-005"])
    genome_pair.compare_attribute("host_genus", expect_same=True,
                                  eval_id="GP-EVAL-006", fail="warning",
                                  eval_def=EDD["GP-EVAL-006"])
    genome_pair.compare_attribute("annotation_author", expect_same=True,
                                  eval_id="GP-EVAL-007", fail="warning",
                                  eval_def=EDD["GP-EVAL-007"])
    genome_pair.compare_attribute("translation_table", expect_same=True,
                                  eval_id="GP-EVAL-008", fail="warning",
                                  eval_def=EDD["GP-EVAL-008"])
    genome_pair.compare_attribute("retrieve_record", expect_same=True,
                                  eval_id="GP-EVAL-009", fail="warning",
                                  eval_def=EDD["GP-EVAL-009"])

    if eval_flags["check_replace"]:
        # The following checks assume that:
        # 'genome1' slot = new genome to be evaluated
        # 'genome2' slot = the current MySQL genome

        # The new genome to be evaluated is expected to be
        # newer than the current MySQL genome annotations.
        genome_pair.compare_date("newer", eval_id="GP-EVAL-010", fail="warning",
                                 eval_def=EDD["GP-EVAL-010"])

        if genome_pair.genome2.annotation_status == "draft":
            # It is expected that the replacing genome name no longer
            # retains the "_Draft" suffix, so the name should change.
            genome_pair.compare_attribute("name", expect_same=False,
                                          eval_id="GP-EVAL-011", fail="warning",
                                          eval_def=EDD["GP-EVAL-011"])

            # The status should change from 'draft'.
            genome_pair.compare_attribute("annotation_status", expect_same=False,
                                          eval_id="GP-EVAL-012", fail="warning",
                                          eval_def=EDD["GP-EVAL-012"])
        else:
            genome_pair.compare_attribute("name", expect_same=True,
                                          eval_id="GP-EVAL-013", fail="warning",
                                          eval_def=EDD["GP-EVAL-013"])

            # This is tricky. Yes, if replacing, you only expect a
            # final -> final, or unknown -> unknown. However, if the
            # check_replace = True, such as when a new final is available from
            # PhagesDB, then you only expect to go from draft -> final,
            # and you don't expect the current MySQL genome to be final.
            genome_pair.compare_attribute("annotation_status", expect_same=True,
                                          eval_id="GP-EVAL-014", fail="warning",
                                          eval_def=EDD["GP-EVAL-014"])
            genome_pair.compare_attribute("accession", expect_same=True,
                                          eval_id="GP-EVAL-015", fail="warning",
                                          eval_def=EDD["GP-EVAL-015"])


# TODO unit test.
def check_trna(trna_ftr, eval_flags):
    """Check a Trna object for errors.

    :param trna_ftr: A pdm_utils Trna object.
    :type trna_ftr: Trna
    :param eval_flags: Dictionary of boolean evaluation flags.
    :type eval_flags: dicts
    """
    logger.info(f"Checking tRNA feature: {trna_ftr.locus_tag} ({trna_ftr.id}).")

    # Start can be 0 since coordinate_format = 0_half_open
    trna_ftr.check_magnitude("start", ">", -1, eval_id="TRNA-EVAL-001",
                             eval_def=EDD["TRNA-EVAL-001"])
    trna_ftr.check_magnitude("stop", ">", 0, eval_id="TRNA-EVAL-002",
                             eval_def=EDD["TRNA-EVAL-002"])
    trna_ftr.check_magnitude("parts", ">", 0, eval_id="TRNA-EVAL-003",
                             eval_def=EDD["TRNA-EVAL-003"])
    trna_ftr.check_orientation(eval_id="TRNA-EVAL-004",
                               eval_def=EDD["TRNA-EVAL-004"])

    if eval_flags["check_locus_tag"]:
        trna_ftr.check_attribute(
            "locus_tag", {""}, expect=False, eval_id="TRNA-EVAL-005",
            fail="warning", eval_def=EDD["TRNA-EVAL-005"])

        # TODO this check could be improved to take into account the prefix.
        trna_ftr.check_locus_tag_structure(
            check_value=None, only_typo=True, case=True, eval_id="TRNA-EVAL-006",
            fail="warning", eval_def=EDD["TRNA-EVAL-006"])

    if eval_flags["check_gene"]:
        trna_ftr.check_attribute(
            "gene", {""}, expect=False, eval_id="TRNA-EVAL-007", fail="warning",
            eval_def=EDD["TRNA-EVAL-007"])
        trna_ftr.check_gene_structure(
            eval_id="TRNA-EVAL-008", fail="warning",
            eval_def=EDD["TRNA-EVAL-008"])

    if eval_flags["check_locus_tag"] and eval_flags["check_gene"]:
        trna_ftr.check_compatible_gene_and_locus_tag(
            eval_id="TRNA-EVAL-009", fail="warning",
            eval_def=EDD["TRNA-EVAL-009"])

    # All tRNAs should be checked for a valid isotype and anticodon
    trna_ftr.check_amino_acid_valid(
        eval_id="TRNA-EVAL-010", eval_def=EDD["TRNA-EVAL-010"])
    trna_ftr.check_anticodon_valid(
        eval_id="TRNA-EVAL-012", eval_def=EDD["TRNA-EVAL-012"])

    # if check_trna flag is set, we should run Aragorn and tRNAscan-SE to
    # validate the correctness of the annotation
    if eval_flags["check_trna"]:
        # 1. check_sources
        trna_ftr.check_sources(
            eval_id="TRNA-EVAL-017", fail="warning",
            eval_def=EDD["TRNA-EVAL-017"])
        # 2. check_amino_acid_correct
        trna_ftr.check_amino_acid_correct(
            eval_id="TRNA-EVAL-011", fail="warning",
            eval_def=EDD["TRNA-EVAL-011"])
        # 3. check_anticodon_correct
        trna_ftr.check_anticodon_correct(
            eval_id="TRNA-EVAL-013", fail="warning",
            eval_def=EDD["TRNA-EVAL-013"])
        # 4. check_length
        trna_ftr.check_length(
            eval_id="TRNA-EVAL-014", fail="warning",
            eval_def=EDD["TRNA-EVAL-014"])
        # 5. check_coordinates
        trna_ftr.check_coordinates(
            eval_id="TRNA-EVAL-018", fail="warning",
            eval_def=EDD["TRNA-EVAL-018"])
        # 6. check_orientation_correct
        trna_ftr.check_orientation_correct(
            eval_id="TRNA-EVAL-016", fail="warning",
            eval_def=EDD["TRNA-EVAL-016"])
        # 7. check_terminal_nucleotides
        # trna_ftr.check_terminal_nucleotides(
        #     eval_id="TRNA-EVAL-015", fail="warning",
        #     eval_def=EDD["TRNA-EVAL-015"])


def check_tmrna(tmrna_ftr, eval_flags):
    """Check a Tmrna object for errors.

    :param tmrna_ftr: A pdm_utils Cds object.
    :type tmrna_ftr: Tmrna
    :param eval_flags: Dictionary of boolean evaluation flags.
    :type eval_flags: dicts
    """
    logger.info(f"Checking tRNA feature: {tmrna_ftr.locus_tag} ({tmrna_ftr.id}).")

    # Start can be 0 since coordinate_format = 0_half_open
    tmrna_ftr.check_magnitude("start", ">", -1, eval_id="TMRNA-EVAL-001",
                              eval_def=EDD["TMRNA-EVAL-001"])
    tmrna_ftr.check_magnitude("stop", ">", 0, eval_id="TMRNA-EVAL-002",
                              eval_def=EDD["TMRNA-EVAL-002"])
    tmrna_ftr.check_magnitude("parts", "=", 1, eval_id="TMRNA-EVAL-003",
                              eval_def=EDD["TMRNA-EVAL-003"])
    tmrna_ftr.check_orientation(eval_id="TMRNA-EVAL-004",
                                eval_def=EDD["TMRNA-EVAL-004"])

    if eval_flags["check_locus_tag"]:
        tmrna_ftr.check_attribute(
            "locus_tag", {""}, expect=False, eval_id="TMRNA-EVAL-005",
            fail="warning", eval_def=EDD["TMRNA-EVAL-005"])

        # TODO this check could be improved to take into account the prefix.
        tmrna_ftr.check_locus_tag_structure(
            check_value=None, only_typo=True, case=True,
            eval_id="TMRNA-EVAL-006", fail="warning",
            eval_def=EDD["TMRNA-EVAL-006"])

    if eval_flags["check_gene"]:
        tmrna_ftr.check_attribute(
            "gene", {""}, expect=False, eval_id="TMRNA-EVAL-007",
            fail="warning", eval_def=EDD["TMRNA-EVAL-007"])
        tmrna_ftr.check_gene_structure(
            eval_id="TMRNA-EVAL-008", fail="warning",
            eval_def=EDD["TMRNA-EVAL-008"])

    if eval_flags["check_locus_tag"] and eval_flags["check_gene"]:
        tmrna_ftr.check_compatible_gene_and_locus_tag(
            eval_id="TMRNA-EVAL-009", fail="warning",
            eval_def=EDD["TMRNA-EVAL-009"])

    tmrna_ftr.check_peptide_tag_valid(
        eval_id="TMRNA-EVAL-010", eval_def=EDD["TMRNA-EVAL-010"])

    if eval_flags["check_tmrna"]:
        tmrna_ftr.check_peptide_tag_correct(
            eval_id="TMRNA-EVAL-011", fail="warning",
            eval_def=EDD["TMRNA-EVAL-011"])
        tmrna_ftr.check_orientation_correct(
            eval_id="TMRNA-EVAL-012", fail="warning",
            eval_def=EDD["TMRNA-EVAL-012"])


def import_into_db(bndl, engine=None, gnm_key="", prod_run=False):
    """Import data into the MySQL database.

    :param bndl: same as for run_checks().
    :param engine: same as for data_io().
    :param gnm_key:
        Identifier for the Genome object in the Bundle's genome dictionary.
    :type gnm_key: str
    :param prod_run: same as for data_io().
    """
    if bndl._errors == 0:
        import_gnm = bndl.genome_dict[gnm_key]

        # If locus_tag data should not be imported, clear all locus_tag data.
        if bndl.ticket.eval_flags["import_locus_tag"] is False:
            logger.info("Clearing locus_tag data for "
                        f"genome: {import_gnm.id}.")
            import_gnm.clear_locus_tags()

        # Update the date field to reflect the day of import.
        import_gnm.date = IMPORT_DATE
        bndl.sql_statements = mysqldb.create_genome_statements(
                                import_gnm, bndl.ticket.type)
        if prod_run:
            logger.info("Importing data into the database for "
                        f"genome: {import_gnm.id}.")
            execute_result, msg = mysqldb.execute_transaction(engine,
                                        bndl.sql_statements)
            if execute_result == 1:
                result = False
                logger.error("Error importing data. " + msg)
            else:
                result = True
                logger.info("Data successfully imported. " + msg)
                logger.info("The following MySQL statements were executed:")
                for statement in bndl.sql_statements:
                    statement = basic.truncate_value(statement, 150, "...")
                    logger.info(statement)

            # Result of statement execution is stored in an evaluation object
            # so that it can be recorded in the log file.
            # It is stored in the bundle object instead of the genome object
            # since the collection of MySQL statements are constructed
            # based on the ticket data and the genome data combined,
            # and not just the genome data.
            bndl.check_statements(execute_result, msg, eval_id="BNDL-EVAL-007",
                                  eval_def=EDD["BNDL-EVAL-007"])
        else:
            result = True
            logger.info("Data can be imported if set to production run.")
    else:
        result = False
        logger.info("Data contains errors, so it will not be imported.")

    return result



import re
from collections import deque
import sys
import tempfile

import urllib3
from urllib3.exceptions import (MaxRetryError, PoolError, RequestError)
import yaml
from yaml.scanner import ScannerError

# GLOBAL FUNCTIONS
# -----------------------------------------------------------------------------
POOL_MAX_CONN_ATTEMPTS_MSG = (
                "Maximum connection pool construction attempts reached.\n  "
                "Please check your internet connection.")
RESPONSE_MAX_CONN_ATTEMPTS_MSG = (
                "Maximum request attempts reached.\n\t"
                "Please check the availability at the requested url.")

HREF_NAME_FORMAT = ("""<a href="([-_\w\d./]+)">""")
HREF_DATE_FORMAT = ("""<td align="right">"""
                    """(\d+[-]\d+[-]\d+)\s+\d+[:]\d+\s+</td>""")


# GET FUNCTIONS
# -----------------------------------------------------------------------------
def create_pool(pipeline=False):
    try:
        pool = urllib3.PoolManager()
    except MaxRetryError:
        if pipeline:
            print(POOL_MAX_CONN_ATTEMPTS_MSG)
            sys.exit(1)

        raise PoolError(POOL_MAX_CONN_ATTEMPTS_MSG)

    return pool


def pool_request(url, pool=None,
                 pipeline=False, preload=False, expect_status=200):
    """
    Create a urllib3 PoolManager to access the url link for list of databases
    :returns An HTTPResponse object
    :rtype: urllib3.response.HTTPResponse
    """
    if pool is None:
        pool = create_pool()
    try:
        response = pool.request("GET", url, preload_content=preload)
    except MaxRetryError:
        url_fail_msg = "\n".join([RESPONSE_MAX_CONN_ATTEMPTS_MSG,
                                  "".join(["\t\tRequest failed for ", url])])
        if pipeline:
            print(url_fail_msg)
            sys.exit(1)

        raise RequestError(pool, url, url_fail_msg)

    pool.clear()

    if pipeline:
        if response.status != expect_status:
            print("Received invalid response from server.\n"
                  "Aborting pipeline.")
            sys.exit(1)

    return response


def download_file(file_url, filepath, pool=None, chunk_size=512000,
                  preload=False, verbose=True, expect_status=200):
    """
    Retrieve a file from the server.
    :param file_url:  URL for the desired file:
    :type file_url: str
    :param filepath: Local path where the file can be downloaded to.
    :type filepath: Path
    :returns: If status of the file retrieved from the server is expected.
    :rtype: bool
    """
    request = pool_request(file_url, pool=pool, pipeline=False)
    status = request.status

    if status == expect_status:
        filesize = int(request.getheader("Content-Length"))
        output_file = filepath.open(mode="wb")
        for chunk in request.stream(chunk_size):
            output_file.write(chunk)
            progress = int(request._fp_bytes_read / filesize * 100)
            if verbose:
                print("\r[{}{}] {}%".format("#" * int(progress / 2),
                                            " " * (50 - int(progress / 2)),
                                            progress), end="")
    else:
        if verbose:
            print(" ".join(["ERROR.  HTTP Response", str(status)]))

    if verbose:
        print("")

    return status == expect_status


def spool_file(file_url, pool=None, max_size=512000, chunk_size=512000,
               expect_status=200):
    """
    Retrieve a file from the server.
    :param file_url:  URL for the desired file:
    :type file_url: str
    :param filepath: Local path where the file can be downloaded to.
    :type filepath: Path
    :param expect_status: Status expected from the HTTPresponse object
    :type expect_status: int
    :returns: Returns temporary file handle with the downloaded data
    :rtype: tempfile.SpooledTemporaryFile
    """
    request = pool_request(file_url, pool=pool, pipeline=False)
    status = request.status

    if status == expect_status:
        spooled_file = tempfile.SpooledTemporaryFile(max_size=max_size)
        for chunk in request.stream(chunk_size):
            spooled_file.write(chunk)

    return spooled_file


def get_yaml_data(response):
    """Parses yaml data from a response using the PyYAML library.

    :param response: An HTTPrepsonse object:
    :type response: urllib3.response.HTTPResponse
    :return: Dictionary containing key-value pairs from the YAML file
    :rtype: dict
    """
    try:
        data = yaml.safe_load(response.data)
    except ScannerError:
        data = None

    return data


def get_url_listing_data(response, get_dates=False):
    """
    Get list of file and directory data from the link using a response object
    :param response: PoolManager response for the specified url link
    :type response: urllib3.response.HTTPResponse
    :returns: A list of entry listing data
    :rtype: list[dict]
    """
    listings = list()

    response_data = response.data.decode("utf-8")
    response_data_lines = response_data.split("\n")

    # Get database names
    names_regex = re.compile(HREF_NAME_FORMAT)
    dates_regex = re.compile(HREF_DATE_FORMAT)

    data = []
    for line in response_data_lines:
        name_match = re.match(names_regex, line)
        date_match = re.match(dates_regex, line)

        date = None
        if date_match is not None:
            # Retrieves the first capturing group match to the date regex
            date = date_match[1]

        name = None
        if name_match is not None:
            # Retrieves the first capturing group match to the name regex
            name = name_match[1]

        if name is not None:
            data.append((name, date))

    response.close()

    # creating a list of dictionaries
    for i in range(len(data)):
        db_dict = dict()
        db_dict["num"] = i + 1
        db_dict["name"] = data[i][0]
        db_dict["date"] = data[i][1]

        listings.append(db_dict)

    return listings


# PARSING FUNCTIONIS
# ----------------------------------------------------------------------------
def get_url_listing_names(response, suffix=None, ignore=None):
    """Get a list of file and directory names at the url using a HTTPresponse

    :param response: PoolManager response for the specified url link
    :type response: urllib3.response.HTTPResponse
    :param suffix: Suffix string to filter names for
    :type suffix: str
    :param ignore: Prefix string to filter names against
    :type ignore: str
    :returns: A list of entry listing names
    :rtype: list
    """
    listing = get_url_listing_data(response)

    listing_names = list()

    for data_dict in listing:
        name = data_dict.get("name")

        if name is None:
            continue

        if ignore:
            if name.startswith(ignore):
                continue

        if suffix:
            if len(suffix) >= len(name):
                continue

            if name.endswith(suffix):
                # Removes suffix before appending to names list
                listing_names.append(name[:-(len(suffix))])
        else:
            listing_names.append(name)

    return listing_names


def get_url_listing_dirs(response, ignore="."):
    """Get a list of file and directory names at the url using a HTTPresponse

    :param response: PoolManager response for the specified url link
    :type response: urllib3.response.HTTPResponse
    :param ignore: Prefix string to filter names against
    :type ignore: str
    :returns: A list of entry listing names
    :rtype: list
    """
    return get_url_listing_names(response, suffix="/", ignore=ignore)


def get_url_listing_files(response, file_ext, ignore=None):
    """Get a list of file and directory names at the url using a HTTPresponse

    :param response: PoolManager response for the specified url link
    :type response: urllib3.response.HTTPResponse
    :param file_ext: File extension to filter names for
    :type file_ext: str
    :param ignore: Prefix string to filter names against
    :type ignore: str
    :returns: A list of entry listing names
    :rtype: list
    """
    return get_url_listing_names(response, "".join([".", str(file_ext)]),
                                 ignore=ignore)


class UrlPathGraph():
    """Graph object that manages and stores PathNodes representing
    directories at an url"""
    def __init__(self, url, pool=None, pipeline=False, preload=False,
                 expect_status=200):
        self.nodes = list()

        if pool is None:
            pool = create_pool(pipeline=pipeline)

        self.pool = pool
        self.expect_status = expect_status
        self.pipeline = pipeline

        self.url = url
        self.root = PathNode("", root_path=url, isdir=True)
        self.curr = self.root
        self.nodes.append(self.root)
        self.load_node(self.root)

        if preload:
            self.load_graph()

    def load_graph(self):
        """Loads the structure of a filesystem at the specified url as a graph
        """
        node_stack = deque()

        # Do/while loop that iterates until all directories are explored
        parent = self.root
        while True:
            if not parent.loaded:
                self.load_node(parent)

            for child_name, child_node in parent.children.items():
                # Ignores hidden nodes
                if not child_name.startswith("."):
                    node_stack.append(child_node)

            try:
                parent = node_stack.pop()
            except IndexError:
                break

    def load_node(self, node):
        """Uses a HTTPresponse object to gather data to load a PathNode

        :param node: PathNode to load data and children for
        :type node: PathNode
        """
        response = pool_request(node.get_abs_path(), pool=self.pool,
                                expect_status=self.expect_status,
                                pipeline=self.pipeline)

        node.contents = get_url_listing_names(response)
        for child_dir in get_url_listing_dirs(response):
            child_node = PathNode(child_dir, root_path=self.url,
                                  parent=node)

            self.nodes.append(child_node)

            node.children[child_dir] = child_node

        if len(node.children) > 1:
            node.isdir = True
            if node.parent is not None:
                node.children[".."] = node.parent

        response.close()
        node.loaded = True

    def traverse(self, path):
        """ Traverses a linked list using a filesystem path

        :param path: A filesystem path (node names separated by '/')
        :type path: str
        :return: Returns the path node specified by the filesystem path
        :rtype: PathNode
        """
        curr = self.traverse_relative(path, self.curr)

        if curr is None:
            curr = self.traverse_absolute(path)

        return curr

    def traverse_relative(self, path, node):
        """Traverses a linked list using a filesystem path relative to a node

        :param path: A filesystem path (node names separated by '/')
        :type path: str
        :param node: A PathNode designating the start for the path
        ;type node: PathNode
        :return: Returns the path node specified by the filesystem path
        :rtype: PathNode
        """
        node_stops = path.split("/")

        curr = node
        for node_name in node_stops:
            if not curr.loaded:
                self.load_node(curr)

            curr = curr.children.get(node_name)

            if curr is None:
                return None

        return curr

    def traverse_absolute(self, path):
        """Traverses a linked list using a filesystem path relative to root

        :param path: A filesystem path (node names separated by '/')
        :type path: str
        :return: Returns the path node specified by the filesystem path
        :rtype: PathNode
        """
        node_stops = path.split("/")

        if node_stops[0] == "":
            node_stops.pop(0)

        path = "/".join(node_stops)
        return self.traverse_relative(path, self.root)


class PathNode():
    """Object that stores information about a directory in a file system"""
    def __init__(self, name, root_path="", rel_path=None, abs_path=None,
                 isdir=False, children=None, parent=None, data=None):
        self.name = name
        self.isdir = isdir

        self.root_path = root_path
        self.rel_path = rel_path
        self.abs_path = abs_path

        if children is None:
            children = dict()
        children["."] = self

        self.children = children
        self.parent = parent

        self.data = data
        self.contents = []
        self.loaded = False

    def calc_rel_path(self, end_node=None):
        """Calculates the filesystem path relative to the node reference

        :param end_node: Node to create the relative path from
        :type end_node: PathNode
        :return: Returns the path relative to the node reference
        :rtype: str
        """
        path_nodes = list()
        curr_node = self
        while True:
            path_nodes.insert(0, curr_node.name)

            curr_node = curr_node.parent
            if curr_node == end_node or curr_node is None:
                break

        return "/".join(path_nodes)

    def set_rel_path(self, end_node=None):
        """Sets the filesystem path relative to the node reference

        :param end_node: Node to create the relative path from
        :type end_node: PathNode
        """
        self.rel_path = self.calc_rel_path(end_node=end_node)

    def set_abs_path(self):
        """Sets the filesystem path relative to the root directory
        """
        self.abs_path = "/".join([self.root_path.rstrip("/"),
                                  self.calc_rel_path().lstrip("/")])

    def get_rel_path(self, node=None):
        """Retrieves the filesystem path relative to the deepest node reference

        :return: Returns the relative path
        :rtype: str
        """
        if self.rel_path is None:
            self.set_rel_path()

        return self.rel_path

    def get_abs_path(self):
        """Retrieves the filesystem path relative to the root directory

        :return: Returns the absolute path
        :rtype: str
        """
        if self.abs_path is None:
            self.set_abs_path()

        return self.abs_path

    def print_children(self, max_width=74):
        """Prints the filesystem directory child directories
        """
        names = list()
        max_len = 0
        for child_name, child_node in self.children.items():
            if child_name == ".":
                continue

            if child_node.isdir:
                child_name = "".join([child_name, "/"])

            names.append(child_name)
            if len(child_name) > max_len:
                max_len = len(child_name)

        for i in range(len(names)):
            child = names[i]
            child = "".join([child, " " * (max_len - len(child)), "\t"])
            names[i] = child

        cols = int(max_width / (max_len + (4 - max_len % 4)))

        for row in range(0, len(names), cols):
            child_row = names[row:row+cols]
            print("".join(["\t"] + child_row))

    def print_contents(self, max_width=74):
        """Prints the filesystem directory child contents
        """
        names = list()
        max_len = 0
        for name in self.contents:
            names.append(name)

            if len(name) > max_len:
                max_len = len(name)

        for i in range(len(names)):
            name = names[i]
            name = "".join([name, " " * (max_len - len(name)), "\t"])
            names[i] = name

        cols = int(max_width / (max_len + (4 - max_len % 4)))

        for row in range(0, len(names), cols):
            name_row = names[row:row+cols]
            print("".join(["\t"] + name_row))



"""Pipeline that uses rpsblast to search for NCBI conserved domain database
hits in all new proteins in the phage database."""

import argparse
import logging
import pathlib
import shutil
import sys
import tempfile

import sqlalchemy

import pdm_utils
from pdm_utils.classes.alchemyhandler import AlchemyHandler, Engine
from pdm_utils.constants import constants
from pdm_utils.functions import basic, configfile, mysqldb, mysqldb_basic
from pdm_utils.functions import multiprocess as mp, rpsblast


# Miscellaneous
VERSION = pdm_utils.__version__
EVALUE = 0.001
BATCH = 10000
OUTDIR = pathlib.Path().cwd()
RESULTS_FOLDER = f"{constants.CURRENT_DATE}_find_domains"
MAIN_LOG_FILE = "find_domains.log"
DEFAULT_CDD = pathlib.Path().home().joinpath("Databases/Cdd_LE")

# Add a logger named after this module, with a null handler that will suppress
# any output statements. This allows other modules to define the handler and
# output formats. If this module is the main module being called, the top
# level main function configures the root logger and primary file handle.
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# MySQL queries and commands for everything else
GET_NEW_GENES = "SELECT GeneID, Translation FROM gene WHERE DomainStatus = 0"
UPDATE_STATUS = "UPDATE gene SET DomainStatus = 1 WHERE GeneID = '{}'"
INSERT_DOMAIN = """INSERT INTO domain (HitID, DomainID, Name, Description) """ \
                """VALUES ("{}", "{}", "{}", "{}")"""
INSERT_GENE_DOMAIN = """INSERT INTO gene_domain """ \
                     """(GeneID, HitID, Expect, QueryStart, QueryEnd) """ \
                     """VALUES ("{}", "{}", "{}", "{}", "{}")"""


def parse_args(args=None):
    """Parse commandline arguments."""
    # Initialize parser and add arguments
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument("database", type=str,
                        help="name of database to find conserved domains for")

    parser.add_argument("--cdd-dir", type=pathlib.Path, default=DEFAULT_CDD,
                        help=f"path to local NCBI Cdd [default: {DEFAULT_CDD}]")
    parser.add_argument("--cpus", default=mp.CPUS, type=int,
                        help=f"number of CPU cores to use [default: {mp.CPUS}]")
    parser.add_argument("-e", "--evalue", default=EVALUE, type=float,
                        help=f"significance threshold to retain hits "
                             f"[default: {EVALUE}]")
    parser.add_argument("-o", "--output-dir", type=pathlib.Path, default=OUTDIR,
                        help=f"directory where log data can be output "
                             f"[default: {OUTDIR}]")
    parser.add_argument("-b", "--batch-size", default=BATCH, type=int,
                        help=f"number of translations to search at a time "
                             f"[default: {BATCH}]")
    parser.add_argument("-r", "--rpsblast", default=None, type=str,
                        help="path to rpsblast(+) binary")
    parser.add_argument("-x", "--reset", action="store_true",
                        help="reset NCBI Cdd hits for all genes")
    parser.add_argument("-c", "--config", type=pathlib.Path,
                        help="path to a pdm_utils configuration file "
                             "containing MySQL login data")

    if args:
        return parser.parse_args(args)

    return parser.parse_args()


def detect_rpsblast_name():
    """Detect whether rpsblast or rpsblast+ is installed locally.

    :return: name
    """
    for program in ["rpsblast", "rpsblast+"]:
        version = rpsblast.get_version(program)
        if version:
            return program


def learn_cdd_name(cdd_dir):
    """Return the NCBI Cdd 'name' learned by analyzing the filename of
    every file in the Cdd directory.

    :param cdd_dir: directory where Cdd database files live
    :type cdd_dir: pathlib.Path
    :return: cdd_name
    """
    filenames = list()
    for filepath in cdd_dir.iterdir():
        if filepath.is_file() and filepath.name != ".DS_Store":
            filenames.append(filepath.stem)

    name_set = set(filenames)
    if len(name_set) == 1:
        return filenames[0]


def find_domains(program, sequences, cdd, evalue, cpus):
    """Find and return NCBI conserved domain database hits for the
    given sequences.

    :param program: name (or path) to the program to run (e.g. 'rpsblast')
    :type program: str or pathlib.Path
    :param sequences: the sequences to search for Cdd hits in
    :type sequences: list[tuple[str, str]]
    :param cdd: path to a local copy of the Cdd database
    :type cdd: str or pathlib.Path
    :param evalue: significance threshold to retain hits
    :type evalue: float
    :param cpus: number of cpu cores to use
    :type cpus: int
    :return: domains
    """
    # Calling code needs to clean up temp dir created this way - see
    # try/finally block below
    tmp_dir = tempfile.mkdtemp(prefix="pdm_utils__find_domains")

    try:
        outfiles = rpsblast.blastall(program=program, sequences=sequences,
                                     db=cdd, tmp_dir=pathlib.Path(tmp_dir),
                                     evalue=evalue, cpus=cpus)
        domains = dict()
        for outfile in outfiles:
            temp_domains = rpsblast.parse_xml(outfile, sequences, evalue)
            for geneid, gene_domains in temp_domains.items():
                domains[geneid] = gene_domains
    finally:
        shutil.rmtree(tmp_dir)

    return domains


def main(argument_list):
    """Commandline interface for this pipeline.

    :param argument_list: the list of CLI args given for this pipeline
    :type argument_list: list[str]
    """
    # Parse commandline arguments
    args = parse_args(argument_list[2:])

    # Store arguments in more easily accessible variables
    database = args.database
    cdd_dir = args.cdd_dir
    cdd_name = learn_cdd_name(cdd_dir)
    cpus = args.cpus
    evalue = args.evalue
    program = args.rpsblast
    output_dir = args.output_dir
    reset = args.reset
    batch_size = args.batch_size

    # Create config object with data obtained from file and/or defaults.
    config = configfile.build_complete_config(args.config)
    mysql_creds = config["mysql"]

    # Set up directory.
    output_folder = basic.set_path(output_dir, kind="dir", expect=True)
    results_folder = pathlib.Path(RESULTS_FOLDER)
    results_path = basic.make_new_dir(output_folder, results_folder,
                                      attempt=50)
    if results_path is None:
        print("Unable to create output_folder.")
        sys.exit(1)

    log_file = pathlib.Path(results_path, MAIN_LOG_FILE)

    # Set up root logger.
    logging.basicConfig(
        filename=log_file, filemode="w", level=logging.DEBUG,
        format="pdm_utils find_domains: %(levelname)s: %(message)s")
    logger.info(f"pdm_utils version: {VERSION}")
    logger.info(f"CDD run date: {constants.CURRENT_DATE}")
    logger.info(f"Command line arguments: {' '.join(argument_list)}")
    logger.info(f"Results directory: {results_path}")

    # Early exit if not cdd_name or no rpsblast given and we can't find one
    if not cdd_name:
        msg = (f"Unable to learn CDD database name. Make sure the files in "
               f"{cdd_dir} all have the same basename.")
        logger.error(msg)
        print(msg)
        return

    cdd = cdd_dir.joinpath(cdd_name)

    # If no rpsblast given, try to discover a globally executable version
    if not program:
        program = detect_rpsblast_name()

    if not program:
        msg = "Unable to auto-detect an rpsblast or rpsblast+ executable on " \
              "$PATH. Please install the NCBI-BLAST+ toolkit. If already " \
              "installed, re-run this pipeline with the '--rpsblast' " \
              "argument and provide the path to the rpsblast(+) executable."
        logger.error(msg)
        print(msg)
        return

    # Verify database connection and schema compatibility.
    alchemist = AlchemyHandler(database=database,
                               username=mysql_creds["user"],
                               password=mysql_creds["password"])
    alchemist.connect(pipeline=True)
    engine = alchemist.engine
    logger.info(f"Connected to database: {database}.")
    mysqldb.check_schema_compatibility(engine, "the find_domains pipeline")
    logger.info(f"Schema version is compatible.")
    logger.info("Command line arguments verified.")

    if reset:
        msg = "Resetting all domain data currently in the database."
        logger.info(msg)
        print(msg)
        reset_domain_data(engine)

    # Get gene data that needs to be processed
    # in dict format where key = column name, value = stored value.
    cdd_genes = mysqldb_basic.query_dict_list(engine, GET_NEW_GENES)
    msg = f"{len(cdd_genes)} genes to search for conserved domains..."
    logger.info(msg)
    print(msg)

    # Only run the pipeline if there are genes returned that need it
    if not cdd_genes:
        engine.dispose()
        return

    # Log the GeneIDs we're going to search
    log_gene_ids(cdd_genes)

    # Create dictionary keyed on translation
    nr_dict = dict()
    for gene_dict in cdd_genes:
        geneid = gene_dict["GeneID"]
        translation = gene_dict["Translation"].decode("utf-8")
        if translation in nr_dict:
            nr_dict[translation].append(geneid)
        else:
            nr_dict[translation] = [geneid]
    unique_trans = list(nr_dict.keys())

    msg = (f"{len(unique_trans)} unique translations to search for "
           f"conserved domains...")
    logger.info(msg)
    print(msg)

    # Process translations in batches. Otherwise, searching could take
    # so long that MySQL connection closes resulting in 1 or more
    # transaction errors.
    batch_indices = basic.create_indices(unique_trans, batch_size)
    total_rolled_back = 0
    for start, stop in batch_indices:
        msg = f"Processing translations {start + 1} to {stop}..."
        logger.info(msg)
        print(msg)

        batch_sequences = list()
        for translation in unique_trans[start:stop]:
            batch_sequences.append((nr_dict[translation][0], translation))

        batch_domains = find_domains(program, batch_sequences, cdd,
                                     evalue, cpus)

        batch_transactions = list()
        for translation, domains in batch_domains.items():
            for geneid in nr_dict[translation]:
                batch_transactions.append(create_transaction(geneid, domains))

        batch_rolled_back = insert_domain_data(engine, batch_transactions)
        total_rolled_back += batch_rolled_back

    search_summary(total_rolled_back)
    engine.dispose()


def search_summary(rolled_back):
    """Print search results."""
    if rolled_back > 0:
        msg = (f"Error executing {rolled_back} transaction(s). "
               "Some genes may still contain unidentified domains.")
        logger.error(msg)
    else:
        msg = "All genes successfully searched for conserved domains."
        logger.info(msg)
    print("\n\n\n" + msg)


def create_transaction(geneid, domains):
    """Create a series of MySQL INSERT/UPDATE statements to associate
    a geneid with its newly found conserved domain data.

    :param geneid: the gene to add conserved domain hits for
    :type geneid: str
    :param domains: the conserved domain hit data
    :type domains: list[dict]
    :return: transaction
    """
    transaction = list()

    for hit in domains:
        transaction.append(insert_domain_statement(hit))
        transaction.append(insert_gene_domain_statement(geneid, hit))

    transaction.append(update_gene_status_statement(geneid))

    return transaction


def log_gene_ids(cdd_genes):
    """Record names of the genes processed for reference."""
    batch_indices = basic.create_indices(cdd_genes, 20)
    for indices in batch_indices:
        genes_subset = cdd_genes[indices[0]: indices[1]]
        gene_ids = []
        for gene in genes_subset:
            gene_ids.append(gene["GeneID"])
        logger.info("; ".join(gene_ids))


def insert_domain_statement(domain):
    """Format a SQL statement to insert into the domain table.

    :param domain: a conserved domain hit
    :type domain: dict
    :return: statement
    """
    hit_id, domain_id = domain["HitID"], domain["DomainID"]
    name, description = domain["Name"], domain["Description"]

    if len(name) > 25:
        name = basic.truncate_value(name, 25, "...")

    return INSERT_DOMAIN.format(hit_id, domain_id, name, description)


def insert_gene_domain_statement(geneid, domain):
    """Format a SQL statement to insert into the gene_domain table.

    :param geneid: the gene in which a domain hit was found
    :type geneid: str
    :param domain: a conserved domain hit
    :type domain: dict
    :return: statement
    """
    hit_id, expect = domain["HitID"], domain["Expect"]
    start, end = domain["QueryStart"], domain["QueryEnd"]

    return INSERT_GENE_DOMAIN.format(geneid, hit_id, expect, start, end)


def update_gene_status_statement(geneid):
    """Format a SQL statement to update a gene's domain-search status
    in the gene table.

    :param geneid: a gene for which domain hits were searched
    :type geneid: str
    :return: statement
    """
    """Format the SQL statement to update data in the gene table."""
    return UPDATE_STATUS.format(geneid)


def insert_domain_data(engine, results):
    """Attempt to insert domain data into the database."""
    msg = "Inserting data..."
    logger.info(msg)
    print(msg)

    rolled_back = 0
    connection = engine.connect()
    for result in results:
        exe_result = execute_transaction(connection, result)
        rolled_back += exe_result

    if rolled_back > 0:
        msg = f"Error executing {rolled_back} transaction(s)."
        logger.error(msg)
        print(msg)

    return rolled_back


def reset_domain_data(engine):
    """Reset all domain data stored in the database.

    :param engine: an initialized SQLAlchemy engine
    :type engine: Engine
    :return: failed
    """
    connection = engine.connect()

    transaction = ["TRUNCATE gene_domain", "DELETE FROM domain",
                   "UPDATE gene SET DomainStatus = 0"]

    failed = execute_transaction(connection, transaction)
    if failed:
        logger.error("Unable to clear all domain data.")
    else:
        logger.info("All domain data cleared.")

    return failed


def execute_transaction(connection, statements):
    """Attempt to execute a series of MySQL statements. Failure to
    execute any statement results in rolling back the full transaction.

    :param connection:
    :type connection:
    :param statements:
    :type statements:
    :return: status
    """
    transaction = connection.begin()

    failed = False
    for statement in statements:
        try:
            connection.execute(statement)
        except sqlalchemy.exc.DBAPIError as e:
            pymysql_error_code = e.orig.args[0]
            pymysql_error_message = e.orig.args[1]
            msg = (f"SQLAlchemy Error type: {str(type(e))}. "
                   f"PyMySQL Error type: {str(type(e.orig))}. "
                   f"PyMYSQL Error code: {pymysql_error_code}. "
                   f"PyMySQL Error message: {pymysql_error_message}.")
            if pymysql_error_code == 1062:
                logger.info("Duplicate entry error ignored. " + msg + statement)
            else:
                logger.error("Failed statement execution. " + msg + statement)
                failed = True
                break
        except Exception as e:
            # One cause is that Cdd description contains '%'. Try replacing
            # '%' with '%%' and re-running the statement.
            if "%" in statement:
                logger.warning(f"Failed statement execution ({type(e)}).")
                logger.info("Attempting to resolve '%' error(s).")
                statement = statement.replace("%", "%%")
                try:
                    connection.execute(statement)
                    logger.info("Replacing '%' with '%%' resolved the error.")
                    logger.info("Successful statement execution " + statement)
                except sqlalchemy.exc.DBAPIError as e:
                    pymysql_error_code = e.orig.args[0]
                    if pymysql_error_code == 1062:
                        logger.info("Duplicate entry error ignored. " +
                                    statement)
                except Exception as e:
                    logger.info("Replacing '%' with '%%' did not resolve the "
                                "error.")
                    logger.error(f"Failed statement execution ({type(e)}). " +
                                 statement)
                    failed = True
                    break
            else:
                logger.error(f"Failed statement execution ({type(e)}). " +
                             statement)
                failed = True
                break

    if failed:
        logger.info("Rolling back transaction.")
        transaction.rollback()
    else:
        logger.info("Committing all changes.")
        transaction.commit()

    transaction.close()

    return failed



import shlex
from subprocess import (Popen, DEVNULL)

from pdm_utils.functions import (fileio, multithread, multiprocess)


THREAD_LOCK_TIMEOUT = 10
THREAD_JOIN_TIMEOUT = 30


def run_clustalo(fasta_path, aln_path):
    """Runs Clustal Omega to generate a fasta-formatted  multiple sequence
    alignment file

    :param fasta_path: Path to a fasta file containing sequences to be aligned
    :type fasta_path: Path
    :type fasta_path: str
    :param aln_path: The desired path to the aligned sequences file
    :type aln_path: Path
    :type aln_path: str
    :return: Returns the path to the aligned sequences file
    :rtype: Path
    :rtype: str
    """
    command = (f"clustalo -i {fasta_path} --infmt=fasta -o {aln_path} "
               "--outfmt=fasta --output-order=tree-order --threads=1 "
               "--seqtype=protein")

    command = shlex.split(command)
    with Popen(command, stdout=DEVNULL, stderr=DEVNULL) as proc:
        proc.wait()

    return aln_path


def get_all_pham_gene_translations(alchemist):
    """Creates a 2D dictionary that maps phams to dictionaries that map
    unique translations to respective geneids.

    :param alchemist: A connected and fully built AlchemyHandler object
    :type alchemist: AlchemyHandler
    :return: Returns a dictionary mapping phams to translations to geneids
    :rtype: dict{dict}
    """
    engine = alchemist.engine

    # Build phage>>cluster lookup table
    cluster_lookup = dict()
    host_lookup = dict()
    query = "SELECT PhageID, Cluster, Subcluster, HostGenus FROM phage"
    results = engine.execute(query)
    for result in results:
        phageid = result["PhageID"]
        cluster = result["Cluster"]
        subcluster = result["Subcluster"]
        host = result["HostGenus"]
        if cluster is None:
            cluster_lookup[phageid] = "Singleton"
        elif subcluster is None:
            cluster_lookup[phageid] = cluster
        else:
            cluster_lookup[phageid] = subcluster

        host_lookup[phageid] = host

    # Build pham>>translation>>gene lookup table
    phams = dict()
    query = ("SELECT PhamID, LocusTag, Name, Translation, Notes, PhageID "
             "FROM gene")
    results = engine.execute(query)
    for result in results:
        phageid = result["PhageID"]
        locus = result["LocusTag"]
        translation = result["Translation"].decode('utf-8')
        product = result["Notes"].decode('utf-8')
        phamid = result["PhamID"]
        if locus is not None:
            name = locus.split('_')[-1]
        else:
            name = result["Name"]

        geneid = f"{phageid}_{name}"

        if product is not None and product > "":
            geneid = "".join([geneid, " [product=", str(product), "]"])

        cluster = cluster_lookup[phageid]
        if cluster is not None and cluster > "":
            geneid = "".join([geneid, " [cluster=", str(cluster), "]"])

        host = host_lookup[phageid]
        if host is not None and host > "":
            geneid = "".join([geneid, " [host_genus=", str(host), "]"])

        if locus is not None and locus > "":
            geneid = "".join([geneid, " [locus=", str(locus), "]"])

        pham_translations = phams.get(phamid, dict())
        gene_ids = pham_translations.get(translation, [])
        gene_ids.append(geneid)
        pham_translations[translation] = gene_ids
        phams[phamid] = pham_translations

    engine.dispose()

    return phams


def dump_pham_out_fastas(working_dir, phams_dict, neglect_singles=True,
                         threads=1, verbose=False):
    """Uses multiple threads to write fasta-formatted multiple sequence files
    for all of the phams listed.

    :param working_dir: Path to the directory where the files will be written
    :type working_dir: pathlib.Path
    :param phams_dict: Dictionary that maps phams to translations to geneids
    :type phams_dict: dict{dict}
    :param neglect_singles: Omit single sequence fasta files from the path map
    :type neglect_singles: bool
    :param threads: Number of threads to spawn during export
    :type threads: int
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    :return pham_fasta_map: Dictionary that maps phams to their fasta file path
    :rtype pham_fasta_map: dict
    """
    pham_fasta_map = dict()

    if verbose:
        print("...Writing pham gene fasta files...")

    work_items = []
    for pham, pham_translations in phams_dict.items():
        filename = f"{pham}_genes.fasta"
        filepath = working_dir.joinpath(filename)

        gs_to_ts = {}
        for translation, gene_ids in pham_translations.items():
            gs_to_ts[gene_ids[0]] = translation

        if (not neglect_singles) or (len(gs_to_ts) > 1):
            pham_fasta_map[pham] = filepath

        work_items.append((gs_to_ts, filepath))

    multithread.multithread(work_items, threads, fileio.write_fasta,
                            verbose=verbose, lock_timeout=THREAD_LOCK_TIMEOUT,
                            join_timeout=THREAD_JOIN_TIMEOUT)

    return pham_fasta_map


def align_pham_out_fastas(working_dir, pham_fasta_map, threads=1,
                          verbose=False):
    """Uses multiple processes to align fasta-formatted multiple sequence files
    for all of the phams listed.

    :param working_dir: Path to the directory where the files will be written
    :type working_dir: pathlib.Path
    :param phams_dict: Dictionary that maps phams to their fasta file path
    :type phams_dict: dict{Path}
    :param threads: Number of processes/threads to spawn during alignment
    :type threads: int
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    :return pham_aln_map: Dictionary that maps phams to their aln file path
    :rtype pham_aln_map: dict
    """
    pham_aln_map = dict()

    if verbose:
        print("...Aligning pham gene fasta files...")

    work_items = []
    for pham, filepath in pham_fasta_map.items():
        aln_name = filepath.with_suffix(".aln").name
        aln_path = working_dir.joinpath(aln_name)

        pham_aln_map[pham] = aln_path

        work_items.append((filepath, aln_path))

    multiprocess.parallelize(work_items, threads, run_clustalo,
                             verbose=verbose)

    return pham_aln_map


def reintroduce_pham_fasta_duplicates(pham_path_map, pham_translations_dict,
                                      threads=1, verbose=False):
    """Uses multiple threads to read and re-write fasta-formatted sequence
    files for all of the phams listed.

    :param pham_path_map: Dictionary that maps phams to their file path
    :type pham_path_map: dict{Path}
    :param phams_translations_dict: Map of phams to translations to geneids
    :type phams_translations_dict: dict{dict}
    :param threads: Number of threads to spawn during reading/writing of files
    :type threads: int
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    """
    work_items = []
    for pham, filepath in pham_path_map.items():
        ts_to_gs = pham_translations_dict.get(pham, {})

        work_items.append((ts_to_gs, filepath))

    multithread.multithread(
                        work_items, threads,
                        fileio.reintroduce_fasta_duplicates, verbose=verbose,
                        lock_timeout=THREAD_LOCK_TIMEOUT,
                        join_timeout=THREAD_JOIN_TIMEOUT)


# Parallelized all-encompassing function that combines the functionality
# of the above functions
def write_phams(fasta_dir, aln_dir, phams_translations_dict, cores=1,
                verbose=False):
    work_items = []
    for pham, pham_translations in phams_translations_dict.items():
        work_items.append((fasta_dir, aln_dir, pham, pham_translations))

    multiprocess.parallelize(work_items, cores, write_phams_process,
                             verbose=verbose)


def write_phams_process(fasta_dir, aln_dir, pham, pham_translations):
    fasta_path = fasta_dir.joinpath("".join([str(pham), "_genes.fasta"]))
    aln_path = aln_dir.joinpath("".join([str(pham), "_genes.aln"]))

    gs_to_ts = {}
    for translation, gene_ids in pham_translations.items():
        gs_to_ts[gene_ids[0]] = translation

    fileio.write_fasta(gs_to_ts, fasta_path)

    if len(pham_translations) > 1:
        run_clustalo(fasta_path, aln_path)
        fileio.reintroduce_fasta_duplicates(pham_translations, aln_path)

    fileio.reintroduce_fasta_duplicates(pham_translations, fasta_path)



from datetime import datetime
from decimal import Decimal
import re

from sqlalchemy import Column
#----------------------------------------------------------------------------
#GLOBAL VARIABLES
VALUE_CHARACTERS      = "\w\d\-_%/(),'"
VC = VALUE_CHARACTERS
VALUE_EXPRESSION      = f"'[ {VC}]*'|[{VC}]+"
VE = VALUE_EXPRESSION
FILTER_FORMAT         = (" *\w+\.\w+ *([=<>!]+ *| +LIKE +| +IS NOT +){1} *"
                        f"({VE}) *")

IN_FORMAT             = (" *\w+\.\w+ *( +IN +| +NOT IN +)"
                        f"\(([\s\w\W\d]+)\)")



NUMERIC_OPERATORS     = [">", ">=", "<", "<="]
NONNUMERIC_OPERATORS  = ["=", "!=", "IS NOT", "LIKE", "IN", "NOT IN"]
OPERATORS             = NUMERIC_OPERATORS + NONNUMERIC_OPERATORS
COMPARABLE_TYPES      = [int, Decimal, float, datetime]
TYPES                 = [str, bytes] + COMPARABLE_TYPES
GROUP_OPTIONS         = ["limited_set", "num_set", "str_set"]

#----------------------------------------------------------------------------
#SPACE HANDLING

def parse_out_ends(unparsed_string):
    """Parse and remove beginning and end whitespace of a string.

    :param unparsed_string: String with variable terminal whitespaces.
    :type unparsed_string: str
    :returns: String with parsed and removed beginning and ending whitespace.
    :rtype: str
    """
    if not isinstance(unparsed_string, str):
        raise TypeError(f"Parameter type required is of type list.")

    beginning_trimmed = unparsed_string.lstrip()
    trimmed_string = beginning_trimmed.rstrip()

    return trimmed_string

def parse_in_spaces(unparsed_string_list):
    """Convert a list of strings to a single space separated string.

    :param unparsed_string_list: String list to be concatenated
    :type unparsed_string_list: list[str]
    :returns: String with parsed in whitespace.
    :rtype: str
    """
    if not isinstance(unparsed_string_list, list):
        raise TypeError(f"Parameter type required is of type list.")

    joined_string = unparsed_string_list[0]
    unparsed_string_list = unparsed_string_list[1:]

    for string in unparsed_string_list:
        joined_string = (f"{joined_string} {string}")

    return joined_string

def parse_out_spaces(unparsed_string):
    """Parse and remove beginning and internal white space of a string.

    :param unparsed_string: String with variable terminal whitespaces.
    :type unparsed_string: str
    :returns: String with parsed and removed beginning and ending whitespace.
    :rtype: str
    """
    if not isinstance(unparsed_string, str):
        raise TypeError(f"Paremeter type required is of type list.")

    trimmed_string = parse_out_ends(unparsed_string)
    split_string = unparsed_string.split(" ")
    joined_string = "".join(split_string)

    return joined_string

#----------------------------------------------------------------------------
#MYSQL TABLE/COLUMN HANDLING

def parse_column(unparsed_column):
    """Recognizes and parses a MySQL structured column.

    :param unparsed_column: Formatted MySQL column.
    :type unparsed_column: str
    :returns: List containing segments of a MySQL column.
    :rtype: list[str]
    """
    column_format = re.compile(" *\w+\.\w+ *", re.IGNORECASE)

    if re.match(column_format, unparsed_column) != None:
        trimmed_column = parse_out_ends(unparsed_column)
        parsed_column = re.split("\.", trimmed_column)

        parsed_column[0] = parse_out_ends(parsed_column[0])
        parsed_column[1] = parse_out_ends(parsed_column[1])
    else:
        raise ValueError(f"Unsupported table/column format: "
                         f"'{unparsed_column}'")

    return parsed_column

def translate_table(metadata, raw_table):
    """Converts a case-insensitive table name to a case-sensitive str.

    :param metadata: Reflected SQLAlchemy MetaData object.
    :type metadata: MetaData
    :param raw_table: Case-insensitive table name.
    :type_table: str
    :returns: Case-sensitive table name.
    :rtype: str
    """
    for table in metadata.tables.keys():
        if table.lower() == raw_table.lower():
            return table

    raise ValueError(f"Table '{raw_table}' requested to be filtered "
                     f"is not in selected database.")

def translate_column(metadata, raw_column):
    """Converts a case-insensitve {table}.{column} str to a case-sensitive str.

    :param metadata: Reflected SQLAlchemy MetaData object.
    :type metadata: MetaData
    :param raw_column: Case-insensitive {table}.{column}.
    :type raw_column: str
    :returns: Case-sensitive column name.
    :rtype: str
    """
    parsed_column = parse_column(raw_column)

    table = translate_table(metadata, parsed_column[0])
    table_obj = metadata.tables[table]
    for column in table_obj.columns.keys():
        if column.lower() == parsed_column[1].lower():
            return column

    raise ValueError(f"Field '{parsed_column[1]}' requested to be filtered"
                     f" is not in '{table_obj.name}'")

#----------------------------------------------------------------------------
#MYSQL CONDITIONAL STRING HANDLING

def parse_filter(unparsed_filter):
    """Recognizes and parses a MySQL structured WHERE clause.

    :param unparsed_filter: Formatted MySQL WHERE clause.
    :type unparsed_filters: str
    :returns: List containing segments of a MySQL WHERE clause.
    :rtype: list[str]
    """

    filter_format = re.compile(FILTER_FORMAT) 
    in_format = re.compile(IN_FORMAT) 

    value_format = re.compile(f"({VE})")
    quote_value_format = re.compile(f"('[ {VC}]*'|'[ {VC}]+'s')")
    in_delimiter = re.compile("' *,")
    whitespace = re.compile(" +")

    if not re.match(filter_format, unparsed_filter) is None:
        operators = re.compile("( *[=<>!]+ *| +LIKE +| +IS NOT +)")

        operator_split = re.split(operators, unparsed_filter)
        parsed_column = parse_column(operator_split[0])
       
        operator_split[1] = parse_out_ends(operator_split[1])
        operator_split[2] = parse_out_ends(operator_split[2])

        if re.match(quote_value_format, operator_split[2]) != None:
            operator_split[2] = operator_split[2].rstrip("'")
            operator_split[2] = operator_split[2].lstrip("'")

        parsed_filter = parsed_column + operator_split[1:] 

    elif not re.match(in_format, unparsed_filter) is None:
        operators = re.compile("( +IN +| +NOT IN +)")

        operator_split = re.split(operators, unparsed_filter)
        parsed_column = parse_column(operator_split[0])

        operator_split[1] = parse_out_ends(operator_split[1]) 
        operator_split[2] = parse_out_ends(operator_split[2])  
 
        parenthesized = operator_split[2]
        parenthesized = parenthesized.lstrip("(")
        parenthesized = parenthesized.rstrip(")")

        in_values = []
        if not re.search(in_delimiter, parenthesized) is None:
            parenthesized = re.split(in_delimiter, parenthesized)     
            for value in parenthesized:
                value = parse_out_ends(value)
                value = value.lstrip("'")
                value = value.rstrip("'")
                in_values.append(parse_out_ends(value))
        else:
            parenthesized = parenthesized.split(",")
            for value in parenthesized:
                value = parse_out_ends(value)
                if value == "" or not re.match(whitespace, value) is None:
                    continue
                in_values.append(value)

        operator_split[2] = in_values
        parsed_filter = parsed_column + operator_split[1:]

    else:
        raise ValueError(f"Unsupported filtering format: '{unparsed_filter}'")
     
    return parsed_filter

def create_filter_key(unparsed_filter):
    """Creates a standardized filter string from a valid unparsed_filter.

    :param unparsed_filter: Formatted MySQL WHERE clause.
    :type unparsed_filters: str
    :returns: Standardized MySQL conditional string.
    :rtype: str
    """
    parsed_filter = parse_filter(unparsed_filter)
    column = ".".join(parsed_filter[0:2])
    
    if isinstance(parsed_filter[3], list):
        parsed_filter[3] = ",".join(parsed_filter[3])
        parsed_filter[3] = "".join(["(", parsed_filter[3], ")"])
    filter_right = "".join(parsed_filter[2:])

    filter_key = column + filter_right
    return filter_key

def check_operator(operator, column_object):
    """Validates an operator's application on a MySQL column.

    :param operator: Accepted MySQL operator.
    :type operator: str
    :param column_object: A SQLAlchemy Column object.
    :type column_object: Column
    """
    if operator not in OPERATORS:
        raise ValueError(f"Operator {operator} is not supported.")
    
    if not isinstance(column_object, Column):
        raise TypeError(f"Type {type(column_object)} is not a Column.")

    column_type = column_object.type.python_type

    if column_type not in TYPES:
        raise ValueError(f"Column '{column_object.name}' "
                         f"has an unsupported type, {column_type}.")
    if operator in NUMERIC_OPERATORS and \
       column_type not in COMPARABLE_TYPES:
        raise ValueError(f"Column '{column_object.name}' "
                         f"is not comparable with '{operator}'.")

def parse_cmd_string(unparsed_cmd_string): 
    """Recognizes and parses MySQL WHERE clause structures.

    :param unparsed_cmd_string: Formatted MySQL WHERE clause string.
    :type unparsed_cmd_string: str
    :returns: 2-D array containing lists of statements joined by ORs.
    :rtype: list[list]
    """
    cmd_line_format = re.compile(".+ + AND|.+ + OR|.+")
   
    or_splits = []
    if unparsed_cmd_string == "":
        pass
    elif re.match(cmd_line_format, unparsed_cmd_string) != None:
        or_splits = []

        or_split = (re.split(" +OR", unparsed_cmd_string))
        for split in or_split: 
            and_split = re.split(" +AND", split)

            for index in range(len(and_split)):
                 and_split[index] = parse_out_ends(and_split[index])

                 if and_split[index] == "":
                     raise ValueError (f"Cmd filter string contains invalid "
                                        "OR or AND conjuctions: "
                                       f"'{unparsed_cmd_string}'")

            or_splits.append(and_split)
    else:
        raise ValueError(f"Unsupported cmd filter string: "
                         f"'{unparsed_cmd_string}'")

    return or_splits

def parse_cmd_list(unparsed_string_list):
    """Recognizes and parses MySQL WHERE clause structures from cmd lists.

    :param unparsed_string_list: Formatted MySQL WHERE clause arguments.
    :type unparsed_string_list: list[str]
    :returns: 2-D array containing lists of statements joined by ORs.
    :rtype: list[list]
    """
    joined_string = parse_in_spaces(unparsed_string_list)
    parsed_cmd_filters = parse_cmd_string(joined_string)

    return parsed_cmd_filters






"""Represents a structure to directly compare data between two or more CDS
features.
"""

# TODO this class needs to be refactored, with attributes and methods
# simplified. The class is used in the compare pipeline, which has only
# been partially refactored since integrating into pdm_utils.
# It relies on extra Cds object attributes that are NOT present
# in the base Cds class and that are added during compare pipeline.
# So do NOT use this class for anything other than in the compare pipeline
# until it has been properly refactored.

# TODO refactor and test.
class CdsPair:

    # Initialize all attributes:
    def __init__(self):

        self.type = ""
        self.cds1 = None
        self.cds2 = None

        self.different_translation = False
        self.different_start_site = False
        self.different_description = False
        self._total_errors = 0


    # Define all attribute setters:

    # TODO refactor and test.
    def compare_cds(self):

        if self.cds1.orientation == "forward":
            if str(self.cds1.start) != str(self.cds2.start):
                self.different_start_site = True
        elif self.cds1.orientation == "reverse":
            if str(self.cds1.stop) != str(self.cds2.stop):
                self.different_start_site = True
        else:
            pass


        product_description_set = set()
        product_description_set.add(self.cds1.description)
        product_description_set.add(self.cds2.product)


        if len(product_description_set) != 1:
            self.different_description = True

        if self.cds1.translation != self.cds2.translation:
            self.different_translation = True

        # Compute total errors
        # First add all matched feature errors
        if self.different_translation:
            self._total_errors += 1

        if self.different_start_site:
            self._total_errors += 1

        if self.different_description:
            self._total_errors += 1

        # Now add all errors from each individual feature
        # You first compute errors for each individual feature.
        # This step is performed here instead of in the mainline code
        # because you need to wait for the feature matching step
        # after the genome matching step.
        self.cds1.check_for_errors()
        self.cds2.check_for_errors()
        self._total_errors += self.cds1._total_errors
        self._total_errors += self.cds2._total_errors



"""Collection of constants and dictionaries used for maintaining the phage
database."""

from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from datetime import datetime, date

# pdm_utils is compatible with this schema version.
CODE_SCHEMA_VERSION = 10

CURRENT_DATE = date.today().strftime("%Y%m%d")

IMPORT_TABLE_STRUCTURE = {
    "order": [
                    "type",
                    "phage_id",
                    "description_field",
                    "eval_mode",
                    "host_genus",
                    "cluster",
                    "subcluster",
                    "accession",
                    "annotation_author",
                    "retrieve_record",
                    "annotation_status"],
    "required": {
                    "type",
                    "phage_id"},
    "optional": {
                    "description_field",
                    "eval_mode",
                    "host_genus",
                    "cluster",
                    "subcluster",
                    "annotation_status",
                    "annotation_author",
                    "accession",
                    "retrieve_record"},
    "valid_ticket": {
                    "type",
                    "phage_id",
                    "description_field",
                    "eval_mode"},
    "valid_retain": {
                    "host_genus",
                    "cluster",
                    "subcluster",
                    "annotation_author",
                    "annotation_status",
                    "accession",
                    "retrieve_record"},
    "valid_retrieve": {
                    "host_genus",
                    "cluster",
                    "subcluster",
                    "accession"},
    "valid_add": {
                    "host_genus",
                    "cluster",
                    "subcluster",
                    "annotation_author",
                    "annotation_status",
                    "accession",
                    "retrieve_record"},
    "valid_parse": {
                    "host_genus",
                    "accession"},
    "keywords": {"parse", "retrieve", "retain", "none"}
}

NAME_SUFFIX = "_Draft"
ANNOTATION_STATUS_SET = {"draft", "final", "unknown"}
ANNOTATION_AUTHOR_SET = {0, 1}
RETRIEVE_RECORD_SET = {0, 1}
EMPTY_DATE = datetime.strptime("1/1/0001 00:00:00", "%m/%d/%Y %H:%M:%S")

# Some locus tags have "PHIRE", but this should not be the case.
LOCUS_TAG_PREFIX_SET = {"SEA", "PBI"}
EMPTY_GENOME_SEQ = Seq("", IUPAC.ambiguous_dna)
EMPTY_PROTEIN_SEQ = Seq("", IUPAC.protein)

# Common list of values that represent empty or null values.
EMPTY_SET = {"",
             "none",
             "null",
             None,
             "not applicable",
             "na",
             "n/a",
             "0",
             0,
             EMPTY_DATE}

# Set up dna and protein alphabets to verify sequence integrity
DNA_ALPHABET = set(IUPAC.IUPACUnambiguousDNA.letters)
PROTEIN_ALPHABET = set(IUPAC.protein.letters)


# Set of all types of allowable ticket actions using this script.
# Add = add a new genome without removing another.
# Remove = delete a genome without adding another.
# Replace = delete a genome and replace it with another.
# Update = make changes to one or more fields related to a genome
# already present in the database (e.g. HostGenus, Cluster, Subcluster, etc.)
IMPORT_TICKET_TYPE_SET = {"add", "replace"}

# Create set of most common gene description genbank qualifiers.
DESCRIPTION_FIELD_SET = {"product", "note", "function"}

# List of names that represent authors that have control over
# a genome record annotations. This constant is stored as a list
# so that multiple names can be stored, if needed.
AUTHOR_SET = {"hatfull"}


# Settings to access data through the PhagesDB API.
API_PREFIX = "https://phagesdb.org/api/phages/"
API_SUFFIX = "/?format=json"
API_HOST_GENERA = "https://phagesdb.org/api/host_genera/"
API_CLUSTERS = "https://phagesdb.org/api/clusters/"

# API at PhagesDB has you specify how many results to return.
# 1 page at length 7500 will return everything for a few years.
API_SEQUENCED = ("https://phagesdb.org/api/sequenced_phages/"
                 f"?page=1&page_size=7500")

# Auto-annotations from PECAAN
PECAAN_PREFIX = "https://discover.kbrinsgd.org/phameratoroutput/phage/"


# Phage name typo correction dictionary.
# Key = Phage name as it is spelled in the GenBank-formatted record.
# Value = Phage name as it is spelled in PhagesDB and/or the MySQL database, and thus
# how it should be spelled in the import ticket.
# The phage name parsed from the GenBank-formatted record gets
# reassigned this corrected name.
# Reasons for the exceptions are indicated.
PHAGE_ID_DICT = {
    # PhagesDB is unable to handle underscores.
    "ATCC29399B_C": "ATCC29399BC",
    "ATCC29399B_T": "ATCC29399BT",

    # Names are spelled differently in GenBank record compared to
    # the original publication.
    "P100_1": "P100.1",
    "P100_A": "P100A",
    "P14": "P14.4",
    "phiELB20": "ELB20",
    "ZL12": "pZL12",

    # 'LeBron' was changed to 'Bron' by ICTV. They won't change it back.
    "Bron": "LeBron",

    # Inadvertent typos introduced at some point that GenBank won't
    # correct since ICTV uses the typos.
    "BBPiebs31": "BPBiebs31",
    "CaptnMurica": "CapnMurica",
    "Fionnbarth": "Fionnbharth",

    # Case changes implemented by ICTV (and now GenBank).
    "Baka": "BAKA",
    "Billknuckles": "BillKnuckles",
    "Crimd": "CrimD",
    "CJW1": "Cjw1",
    "Deadp": "DeadP",
    "Dlane": "DLane",
    "Dotproduct": "DotProduct",
    "Ericb": "EricB",
    "Gumbie": "GUmbie",
    "Heldan": "HelDan",
    "Jaws": "JAWS",
    "Joedirt": "JoeDirt",
    "Kssjeb": "KSSJEB",
    "Littlee": "LittleE",
    "Macncheese": "MacnCheese",
    "Mrgordo": "MrGordo",
    "Packman": "PackMan",
    "lockley": "Lockley",
    "Rockyhorror": "RockyHorror",
    "Ta17a": "TA17a",

    # Remove host data from phage name
    "ReqiDocB7": "DocB7",
    "ReqiPepy6": "Pepy6",
    "ReqiPoco6": "Poco6",
    "ReqiPine5": "Pine5",
    "vB_ArtM-ArV1": "ArV1",
    "vB_ArS-ArV2": "ArV2",
    "vB_MoxS-ISF9": "ISF9"
    }

# Hatfull lab server info
DB_SERVER = "http://databases.hatfull.org/"



import os
import shlex
from subprocess import Popen
import re

from pdm_utils.functions import basic


class TRNAscanSEHandler:
    def __init__(self, identifier, sequence):
        self.id = identifier
        self.sequence = sequence

        # I/O attributes
        self.temp_dir = "/tmp/trnascanse"
        self.input = os.path.join(self.temp_dir, self.id) + ".fasta"
        self.output = os.path.join(self.temp_dir, self.id) + ".out"
        self.out_str = ""

        self.trna_tally = 0
        self.tmrna_tally = 0

        self.trnas = list()
        self.tmrnas = list()

    def write_fasta(self):
        """
        Writes the search sequence to input file in FASTA format.
        :return:
        """
        if not os.path.exists(self.temp_dir):
            os.makedirs(self.temp_dir)
        with open(self.input, "w") as fh:
            fh.write(f">{self.id}\n{self.sequence}\n")

    def run_trnascanse(self, x=10):
        """
        Set up tRNAscan-SE command, then run it. Explanation of
        arguments:
        :param x: score cutoff for tRNAscan-SE
        :type x: int
        :return:
        """
        command = f"tRNAscan-SE -B -H -qQ --detail -X {x} -o /dev/null "
        command += f"-f {self.output} {self.input}"
        Popen(shlex.split(command)).wait()

    def read_output(self):
        """
        Reads the Aragorn output file and joins the lines into a single
        string which it populates into `out_str`.
        :return:
        """
        with open(self.output, "r") as fh:
            self.out_str = "".join(fh.readlines())

    def parse_trnas(self):
        """
        Searches `out_str` for matches to a regular expression for
        tRNAscan-SE tRNAs.
        :return:
        """
        # values:   start, stop,           length,           amino acid,
        # indices:    0     1                 2                   3
        re_str = "\((\d+)-(\d+)\)\s+Length: (\d+)\s+\w+\s+Type: (\w+)\s+" \
                 "Anticodon: (\w+).*?\s+Seq: (\w+)\s+Str: ([>.<]*)"
        # values:           anticodon,      sequence,     structure
        # indices:             4               5             6
        regex = re.compile(re_str, re.MULTILINE | re.DOTALL)

        trnas = regex.findall(self.out_str)

        # Iterate through tRNAs and create dictionaries for each
        for trna in trnas:
            trna_data = dict()

            # Increment tally
            self.trna_tally += 1

            start, stop = int(trna[0]), int(trna[1])
            # If stop > start, forward orientation
            if start < stop:
                orientation = "forward"
            else:
                orientation = "reverse"
                start, stop = stop, start

            # Coordinates need to be converted from 1-based closed to
            # 0-based half open
            start, stop = basic.reformat_coordinates(
                start, stop, "1_closed", "0_half_open")

            amino_acid = trna[3]
            anticodon = trna[4]
            sequence = trna[5]

            # Convert structure to true dot-bracket notation
            structure = trna[6].replace(">", "(").replace("<", ")")
            # Pad structure with periods to ensure full coverage of sequence
            while len(structure) < len(sequence):
                structure += "."

            trna_data["Orientation"] = orientation
            trna_data["Start"] = start
            trna_data["Stop"] = stop
            trna_data["AminoAcid"] = amino_acid
            trna_data["Anticodon"] = anticodon.lower()
            trna_data["Sequence"] = sequence
            trna_data["Structure"] = structure

            self.trnas.append(trna_data)
